"""
This module is used to create and run Evaluators for different models.

Evaluators allow for rapid model simulating or solving with input variable manipulation
and output variable filtering.

Currently there are four specific evaluators: EvaluatorEP (for EnergyPlus), `EvaluatorEH` (for PyEHub)
`EvaluatorGeneric` (for custom functions), and `AdaptiveSR` (for adaptive sampling).
The Evaluators wrap their respective modeling tools with the evaluator interface.
"""

# Python Core Libraries
import copy
import json
import os
import tempfile
import time
import warnings
from abc import ABC, abstractmethod
from functools import lru_cache, wraps
from pathlib import Path
from typing import Callable, Optional, Sequence, Tuple, Union
from warnings import warn

# External Libraries
import numpy as np
import pandas as pd
import platypus
import yaml
from dask.dataframe import from_pandas
from deprecated.sphinx import deprecated
from pandas import DataFrame as DF
from pyehub.outputter import output_excel
from tqdm.auto import tqdm

# BESOS Imports
import config
import eplus_funcs as eplus
import eppy_funcs as ef
import pyehub_funcs as pf
import sampling
from problem import Problem
from besostypes import PathLike


def _list_cache(
    enabled: bool = config.cache["enabled"],
    size: Optional[int] = config.cache["size"],
    typed: bool = config.cache["typed"],
):
    """Decorator that allows the lru_cache (last recently used cache)
    to work on functions that receive lists"""

    # TODO add complete type hint
    def _list_cache_decorator(f: Callable) -> Callable:
        if enabled:
            cache_f = lru_cache(size, typed)(f)

            @wraps(cache_f)
            def safe_f(individual, *args, **kwargs):
                return cache_f(tuple(individual), *args, **kwargs)

            safe_f.cache_clear = cache_f.cache_clear
            return safe_f
        # add a no-op as the cache clear operation so it does not error if disabled in the config
        f.cache_clear = lambda: None
        return f

    return _list_cache_decorator


class AbstractEvaluator(ABC):
    """
    Base class for Evaluators. This template requires that Evaluators are callable.
    It also gives them the df_apply method and result caching.

    Takes a list of values parameterising a model and return objective/constraint results.

    Evaluates each row in a DataFrame and return the output in another DataFrame
    (optionally including the input) and caches results
    """

    error_mode_options = {"Failfast", "Silent", "Print"}

    def __init__(
        self,
        problem: Problem,
        error_mode: str = "Failfast",
        error_value: tuple = None,
        multi: bool = False,
        progress_bar: bool = True,
    ):
        """

        :param problem: description of the inputs and outputs the evaluator will use
        :param error_mode: One of {'Failfast', 'Silent', 'Print'}.
            Failfast: Stop evaluating as soon as an error is encountered.
            Silent: Evaluation will return the `error_value` for any input values that raise an error.
            Print: Same as silent, but warnings are printed to stderr for any errors.
        :param error_value: The value of the evaluation if an error occurs.
            Incompatible with error_mode='Failfast'. Must be a tuple consisting of
            values for the objectives followed by values for the constraints.
        :param multi: whether or not to use multiprocessing, using this disables cache
        :param progress_bar: whether or not to display a progress bar
        """
        self.problem = problem
        self.error_mode = error_mode
        self.error_value = self._convert_error_mode_from_old_format(error_value)
        self.validate_error_mode()
        self.multi = multi
        # this makes the cache instance-specific instead of class-specific
        if not self.multi:
            self._call = _list_cache()(self._uncached_call)
        self.progress_bar = progress_bar
        self.pbar = None

    def validate_error_mode(self) -> None:
        if self.error_mode not in self.error_mode_options:
            msg = (
                f"Invalid error mode '{self.error_mode}', "
                f"only {self.error_mode_options} are allowed"
            )
            raise ValueError(msg)
        if self.error_mode == "Failfast":
            if self.error_value is None:
                return
            else:
                raise ValueError("error value cannot be set when in Failfast mode")
        # intuit error_value if needed
        if self.error_value is None:
            self.error_value = [
                None,
            ] * (self.problem.num_outputs + self.problem.num_constraints)
        if (
            len(self.error_value)
            != self.problem.num_outputs + self.problem.num_constraints
        ):
            raise ValueError("error value must match outputs and constraints length")
        err_out = list(self.error_value[: self.problem.num_outputs])
        err_constraint = list(self.error_value[self.problem.num_outputs :])

        def default(minimize, value):
            if value is None:
                return float("inf") if minimize else float("-inf")
            else:
                return value

        err_out = [
            default(minimize, value)
            for minimize, value in zip(self.problem.minimize_outputs, err_out)
        ]
        for value in err_constraint:
            if value is None:
                raise NotImplementedError(
                    "Cannot Intuit Constraint Error Values."
                    "If you want this feature, request it at"
                    "https://gitlab.com/energyincities/besos/-/issues"
                )
        self.error_value = err_out + err_constraint

    @abstractmethod
    def eval_single(self, values: Sequence, **kwargs) -> tuple:
        """Returns the objective results for a single list of parameter values.

        :param values: A list of values to set each parameter to,
            in the same order as this evaluator's inputs
        :param kwargs: Any keyword arguments
        :return: a tuple of the objectives and constraints
        """
        pass

    def _uncached_call(self, values: Sequence, **kwargs) -> tuple:
        # This method is a workaround to allow lru_cache to be instance-specific and work on __call__
        try:
            self.validate(values)
            if self.problem.depend_params:
                values = self.problem.overwrite_values(values)
            result = self.eval_single(values, **kwargs)
        except Exception as e:
            if self.error_mode != "Silent":
                msg = ""
                if self.problem.inputs is not None:
                    msg += f'for inputs: {self.problem.names("inputs")}'
                msg += f"problematic values were: {values}"
                warn(msg)
            if self.error_mode == "Failfast":
                raise e
            else:
                # TODO: Move error_value to flat format as well
                result = self.error_value
        return result

    def __call__(self, values: Sequence, **kwargs) -> tuple:
        """Returns the objective results for a single list of parameter values.

        :param values: A list of values to set each parameter to,
            in the same order as this evaluator's inputs
        :return: a tuple of the objectives' results
        """
        # Enables validation and caching in subclasses
        # Redirects calls to evaluate a list of values to the eval_single function.
        # Override eval_single, not this method.
        # Values can be empty to allow evaluating at the current state, but this is not the default behaviour
        # If this is not supported, the validate function of a subclass should reject an empty list
        if self.multi:
            return self._uncached_call(values, **kwargs)
        else:
            return self._call(values, **kwargs)

    def cache_clear(self) -> None:
        """Clears any cached vales of calls to this evaluator.
        This should be called whenever the evaluator's outputs could have changed."""
        if not self.multi:
            self._call.cache_clear()

    def validate(self, values: Sequence) -> None:
        """Takes a list of values and checks that they are a valid input for this evaluator."""
        if len(values) != self.problem.num_inputs:
            raise ValueError(
                f"Wrong number of input values."
                f"{len(values)} provided, {self.problem.num_inputs} expected"
            )
        for p, value in zip(self.problem.inputs, values):
            if not p.validate(value):
                raise ValueError(f"Invalid value {value} for parameter {p}")

    def estimate_time(self, df: DF, processes: int = 4) -> None:
        """Prints out a very rough estimate of the amount of time a job will take to complete.
        Will underestimate smaller sample sets but becomes more accurate as they become larger.

        :param df: a DataFrame where each row represents valid input values for this Evaluator.
        :param processes: amount of cores to use (only relevant when using multiprocessing)
        """
        start = time.time()
        res = self(df.iloc[0])
        end = time.time()

        if self.multi and processes > 1:
            estimate = (end - start) * df.shape[0] / processes
        else:
            estimate = (end - start) * df.shape[0]
        h = np.floor(estimate / 3600)
        m = np.floor((estimate % 3600) / 60)
        s = np.floor((estimate % 3600) % 60)
        print(f"Estimated time for completion: {h} hours, {m} minutes, {s} seconds")

    def df_apply(self, df: DF, keep_input=False, processes: int = 4, **kwargs) -> DF:
        """Applies this evaluator to an entire dataFrame, row by row.

        :param df: a DataFrame where each row represents valid input values for this Evaluator.
        :param keep_input: whether to include the input data in the returned DataFrame
        :param processes: amount of cores to use (only relevant when using multiprocessing)
        :return: Returns a DataFrame with one column containing the results for each objective.
        """
        if self.multi and processes > 1:
            ddf = from_pandas(df, npartitions=processes)
            expected_output_shape = pd.DataFrame(
                columns=list(range(self.problem.num_outputs))
            )
            dask_result = ddf.apply(
                self,
                axis=1,
                meta=expected_output_shape,
                result_type="expand",
                **kwargs,
            )
            result = dask_result.compute()
        else:
            if self.progress_bar:
                self.pbar = tqdm(total=len(df.index), desc="Executing", unit="row")
            result = df.apply(self, axis=1, result_type="expand", **kwargs)
        result = result.rename(
            columns={i: name for i, name in enumerate(self.problem.names("outputs"))},
            errors="raise",
        )
        if keep_input:
            result = pd.concat([df, result], axis=1)
        self.cleanup_pbar()
        return result

    def to_platypus(self) -> platypus.Problem:
        """Converts this evaluator (and the underlying problem) to a platypus compatible format

        :return: A platypus Problem that can optimise over this evaluator
        """
        problem = self.problem.to_platypus()
        problem.function = lambda all_outputs: self.package_for_platypus(
            self(all_outputs)
        )
        return problem

    def package_for_platypus(
        self, all_outputs: tuple
    ) -> Union[Tuple, Tuple[Tuple, Tuple]]:
        if self.problem.num_constraints > 0:
            return (
                all_outputs[: self.problem.num_outputs],
                all_outputs[self.problem.num_outputs :],
            )
        else:
            return all_outputs

    def sample(self, num_samples: int) -> DF:
        inputs = sampling.dist_sampler(sampling.lhs, self.problem, num_samples)
        return self.df_apply(inputs)

    def cleanup_pbar(self) -> None:
        try:
            self.pbar.close()
            self.pbar = None
        except AttributeError:
            pass

    def update_pbar(self):
        """Updates the progress bar, marking one more row as completed."""
        try:
            self.pbar.update(1)
        except AttributeError:
            pass

    # a more detailed type hint is possible, but it is messy,
    # the docstring explains the formats more clearly
    @staticmethod
    def _convert_results_from_old_format(result: Sequence) -> Sequence:
        """detects if results are in the old format, and re-writes them if needed.
        results in the new format are returned unchanged.
        A warning about using the old format will be emitted, if applicable.

        Old format: results have the type:
            Tuple[Tuple[values of objectives], Tuple[values of constraints]]
        New format: results have the type:
            Tuple[values of objectives, values of constraints]

        These formats are also both used for error_value

        :param result: the result in either old or new format
        :return: the result in the new format
        """
        if len(result) == 2:
            objectives, constraints = result
            if (
                isinstance(objectives, Sequence)
                and isinstance(constraints, Sequence)
                and all(
                    isinstance(x, (float, int))
                    for x in list(objectives) + list(constraints)
                )
            ):
                result = tuple(objectives) + tuple(constraints)
                warnings.warn(
                    "Evaluators have changed the format used for storing "
                    "outputs and constraints: "
                    "Old format: (output1, output2, ...), "
                    "(constraint1, constraint2, ...) "
                    "New format: (output1, output2, ..., "
                    "constraint1, constraint2, ...)",
                    FutureWarning,
                )
        return result

    # a type hint is possible, but it is messy,
    # the docstring explains the formats more clearly
    def _convert_error_mode_from_old_format(self, error_value):
        """detects if error_values are in the old format, and re-writes them if needed.
        error_values in the new format are returned unchanged.
        A warning about using the old format will be emitted, if applicable.

        :param error_value: the error value in either the old or new format
        :return: the error mode in the new format
        """
        warning = (
            "Evaluators have changed the format used for error values. "
            "The new format accepts either None or a tuple of values. "
            "See the evaluator docstring for details on the new format. "
            "\nYour error mode has automatically been converted to a new format equivalent."
        )
        if error_value is None:
            return error_value
        if tuple(error_value) == (None, None):
            if self.problem.num_outputs + self.problem.num_constraints == 2:
                return error_value
            else:
                warnings.warn(warning, FutureWarning)
                return [
                    None,
                ] * (self.problem.num_outputs + self.problem.num_constraints)
        if len(error_value) != 2:
            return error_value
        # the length 2 case is ambiguous, since we can have either 2 values
        # to use as the error mode values, or an objectives constraint pair
        # from the old style of input
        # if one of the values is a sequence we have an old style pair
        outputs, constraints = error_value
        if isinstance(outputs, Sequence) or isinstance(constraints, Sequence):
            warnings.warn(warning, FutureWarning)
            if outputs is None:
                outputs = [
                    None,
                ] * self.problem.num_outputs
            outputs = list(outputs)
            if constraints is None:
                constraints = [
                    None,
                ] * self.problem.num_constraints
            constraints = list(constraints)
            error_value = outputs + constraints
        return error_value


class EvaluatorGeneric(AbstractEvaluator):
    """Generic Evaluator

    This evaluator is a wrapper around a evaluation function. Can be useful for
    quick debugging.
    """

    eval_func_format = Callable[[Sequence], Tuple[float, ...]]

    def __init__(
        self,
        evaluation_func: eval_func_format,
        problem: Problem,
        error_mode: str = "Failfast",
        error_value: Sequence = None,
        multi: bool = False,
        progress_bar: bool = True,
    ):
        """
        :param evaluation_func: a function that takes as input an list of values,
                and gives as output a tuple of the objective values for that point in the solution space
        :param problem: description of the inputs and outputs the evaluator will use
        :param multi: whether or not to use multiprocessing, using this disables cache
        :param progress_bar: whether or not to display a progress bar
        """

        super().__init__(
            problem=problem,
            error_mode=error_mode,
            error_value=error_value,
            multi=multi,
            progress_bar=progress_bar,
        )
        self._evaluation_func = evaluation_func

    def eval_single(self, values: Sequence) -> Sequence:
        result = self._evaluation_func(values)
        result = self._convert_results_from_old_format(result)
        self.update_pbar()
        return result


@deprecated(
    version="1.6.0",
    reason="EvaluatorSR has been renamed as EvaluatorGeneric with same functionality.",
)
class EvaluatorSR(EvaluatorGeneric):
    """Surrogate Model Evaluator

    This evaluator has been replaced by EvaluatorGeneric, will be removed in a future release.
    """

    pass  # functionality is exactly the same as EvaluatorGeneric, no subclassing needed.


# TODO: Add an option/subclass that automatically bundles several single variable models into a multiobjective model.
class AdaptiveSR(AbstractEvaluator, ABC):
    """A Template for making adaptive sampling based models compatible with the
    evaluator interface.

    Wraps a user specified model training process
    Evaluates the current model
    Retrains the model on new data
    Records training data
    Clears cache when retrained
    Has a reference evaluator used as a ground-truth
    Optional:
    - Finds the best points to add to the model
    - Update the model without fully retraining
    TODO: make a version that can wrap a scikit-learn pipeline to reduce boilerplate
    TODO: make a version that can bundle multiple single objective models together
    """

    # helper functions provided by AdaptiveSR (Generally avoid editing these, but use them as needed)
    # append_data(X, y)
    # do_infill
    # get_from_reference

    # functions with defaults (These can be removed from this template if you like the defaults)
    # They may depend on some of the optional functions in order to work if using the defaults
    # __init__
    # infill -> get_infill, _update_model
    # update_model

    # optional functions (These will not work unless you implement them)
    # get_infill

    # required functions
    # train
    # eval_single

    tabular = Union[DF, np.array]

    def __init__(
        self,
        reference: AbstractEvaluator = None,
        error_mode: str = "Failfast",
        error_value: Sequence = None,
        multi: bool = False,
    ):
        self.reference: AbstractEvaluator = reference
        super().__init__(
            problem=reference.problem,
            error_mode=error_mode,
            error_value=error_value,
            multi=multi,
        )

        self.model = None
        self.data: DF = pd.DataFrame(
            columns=self.problem.names(parts=["inputs", "outputs", "constraints"])
        )

    @property
    def problem(self):
        return self.reference.problem

    @problem.setter
    def problem(self, value: Problem):
        self.reference.problem = value

    def append_data(self, data: tabular, deduplicate: bool = True) -> None:
        """Adds the X and y data to input_data and output_data respectively

        :param data: a table of training data to store
        :param deduplicate: whether to remove duplicates from the combined DataFrame
        :return:
        """
        self.cache_clear()  # TODO: decide on a consistent way of tracking this
        # can we assume users will only modify the data using this method or will call cache_clear themselves
        new_data = self.problem.to_df(data, ["inputs", "outputs", "constraints"])
        self.data = self.data.append(new_data, ignore_index=True)
        if deduplicate:
            self.data.drop_duplicates(inplace=True)

    def get_infill(self, num_datapoints: int) -> tabular:
        """Generates data that is most likely to improve the model, and can be used for retraining.

        :param num_datapoints: the number of datapoints to generate
        :return: the datapoints generated, in some tabular datastructure
        """
        raise NotImplementedError

    def do_infill(self, data: DF) -> None:
        """Updates the model using the inputs X and outputs y, and stores the added data

        :param data: a table of training data
        :return: None
        """
        old_df = self.data
        df, parts = self.problem.partial_df(
            data, parts=["inputs", "outputs", "constraints"]
        )
        if parts == ["inputs"]:
            outputs: DF = self.get_from_reference(df)
            df = pd.concat([df, outputs], axis=1)
        self.append_data(df)
        if self.model is None:
            self.train()
        else:
            self.update_model(df, old_df)

    def update_model(self, new_data: tabular, old_data: DF = None) -> None:
        """Modifies self.model to incorporate the new data.

        This function should not edit the existing data

        :param new_data: a table of inputs and outputs
        :param old_data: the table of inputs and outputs without the new data
        :return: None
        """
        self.train()

    def infill(self, num_datapoints: int) -> None:
        """Adds num_datapoints samples to the model and updates it.

        :param num_datapoints: number of datapoints to add to the model's training set
        :return: None
        """
        inputs: DF = self.problem.to_df(self.get_infill(num_datapoints), "inputs")
        outputs: DF = self.get_from_reference(inputs)
        self.do_infill(pd.concat([inputs, outputs], axis=1))

    @abstractmethod
    def train(self) -> None:
        """Generates a new model using the stored data, and stores it as self.model"""
        pass

    @abstractmethod
    def eval_single(self, values: Sequence, **kwargs) -> Tuple:
        """Evaluates a single input point

        :param values: The datapoint to evaluate
        :param kwargs: Arbitrary keyword arguments.
        :return: A tuple of the predicted outputs for this datapoint
        """
        pass

    def get_from_reference(self, X: tabular) -> DF:
        """Use the reference evaluator to get the real value of a dataframe of datapoints

        :param X: a table containing the datapoints to evaluate
        :return: a DataFrame containing the results of the datapoints
        """
        df = self.problem.to_df(X, "inputs")
        return self.reference.df_apply(df)


class EvaluatorEP(AbstractEvaluator, ABC):
    """This evaluator uses a Problem to modify a building, and then simulate it.
    It keeps track of the building and the weather file."""

    def __init__(
        self,
        problem: Problem,
        building,
        epw: PathLike = config.files["epw"],
        out_dir: PathLike = config.out_dir,
        err_dir: PathLike = config.err_dir,
        error_mode: str = "Failfast",
        error_value: Optional[Sequence] = None,
        multi: bool = False,
        distributed: bool = False,
        version=None,
        progress_bar: bool = True,
        ep_path: PathLike = None,
        *,
        epw_file: PathLike = None,
    ):
        """

        :param problem: a parametrization of the building and the desired outputs
        :param building: the building that is being simulated.
        :param epw: path to the epw file representing the weather
        :param out_dir: the directory used for files created by the EnergyPlus simulation.
        :param err_dir: the directory where files from a failed run are stored.
        :param error_mode: One of {'Failfast', 'Silent', 'Print'}.
            Failfast: Any error aborts the evaluation.
            Silent: Evaluation will return the `error_value` for any lists of values that raise an error.
            Print: Same as silent, but warnings are printed to stderr for any errors.
        :param error_value: The value of the evaluation if an error occurs. Incompatible with error_mode='Failfast'.
        :param multi: whether or not to use multiprocessing, using this disables cache
        :param distributed: Using this flag will save eplus configuration files into memory.
            It lowers performance, but it enables parallel computing with remote workers
        :param version: Deprecated
        :param progress_bar: whether or not to display a progress bar
        :param epw_file: Deprecated. Use epw instead. Path to the epw file representing the weather.
        """

        super().__init__(
            problem=problem,
            error_mode=error_mode,
            error_value=error_value,
            multi=multi,
            progress_bar=progress_bar,
        )
        if out_dir is None:
            self.out_dir = out_dir
        else:
            self.out_dir = Path(out_dir, config.out_wrapper)
            if not os.path.exists(self.out_dir):
                os.makedirs(self.out_dir)

        self.ep_path = ep_path
        self.err_dir = Path(err_dir, config.error_wrapper)

        self.building = building

        # backwards compatibility for renaming epw_file -> epw
        if epw_file and epw != config.files["epw"]:
            raise ValueError(
                "epw_file and epw cannot be used together. Just use epw, as epw file is deprecated."
            )
        elif epw_file:
            epw = epw_file
            warn(
                FutureWarning(
                    "epw_file has been deprecated and will be removed in the future. Use epw instead.",
                )
            )
        # backwards compatibility for removing version
        if version:
            warnings.warn(
                "the version argument is deprecated for run_building,"
                " and will be removed in the future",
                FutureWarning,
            )
            assert version == eplus.get_idf_version(building), "Incorrect version"

        self.epw = epw
        self.distributed = distributed

        if self.multi:
            self.err_dir = None
            self.out_dir = None

        if self.multi and self.distributed:
            # save the epw to memory. When evaluate is run on a worker,
            # the epw will be moved with the object
            with open(self.epw, "r") as f:
                self.epw_ = f.read()
                f.close()

    def df_apply(
        self,
        df: DF,
        keep_input=False,
        processes: int = 2,
        keep_dirs: bool = False,
        *,
        out_dir=config.out_dir,
    ) -> DF:
        """Applies this evaluator to an entire dataFrame, row by row.

        :param df: a DataFrame where each row represents valid input values for this Evaluator.
        :param keep_input: whether to include the input data in the returned DataFrame
        :param processes: amount of cores to use (only relevant when using multiprocessing)
        :param keep_dirs: whether or not keep output directory
        :return: Returns a DataFrame with one column containing the results for each objective.
        """
        result = super().df_apply(
            df,
            keep_input=False,
            processes=processes,
            keep_dirs=keep_dirs,
            out_dir=out_dir,
        )
        # Fixes Dataframe output when using Multiply or Add mode
        for j, parameter in enumerate(self.problem.inputs):
            if not isinstance(parameter, int):
                # FIXME: This code block looks broken
                if parameter.mode != "Set":
                    names = self.problem.names("inputs")
                    if parameter.mode != "Multiply":
                        df.rename(
                            columns={names[j]: names[j] + " Multiplier"},
                            inplace=True,
                            errors="raise",
                        )
                    if parameter.mode != "Add":
                        df.rename(
                            columns={names[j]: names[j] + " Addend"},
                            inplace=True,
                            errors="raise",
                        )
        if keep_input:
            result = pd.concat([df, result], axis=1)
        # TODO: Rework automatic column naming
        if keep_dirs:
            result.rename(
                columns={result.columns[-1]: "output_dir"}, inplace=True, errors="raise"
            )
        self.cleanup_pbar()
        return result

    def save_epw_to_temp(self):
        """Used when evaluate is called and distributed is True
        Writes epw to a temp dir
        Remember to call temp_dir.cleanup() after using the dir"""
        temp_dir = tempfile.TemporaryDirectory()
        temp_epw = Path(temp_dir.name, "temp_epw.epw")

        with open(temp_epw, "w") as f:
            f.write(self.epw_)
            f.close()

        return temp_epw, temp_dir

    def eval_single(self, values: list, out_dir=None, keep_dirs=False):
        out_dir = out_dir or self.out_dir
        err_dir = self.err_dir
        if keep_dirs:
            err_dir = out_dir = ef.generate_dir(out_dir)
        current_building = copy.deepcopy(self.building)
        for p, value in zip(self.problem.inputs, values):
            # apply the modification of each parameter to the building using the corresponding value
            p.transformation_function(current_building, value)
        if self.multi and self.distributed:
            # the remote worker can't access the original epw file
            # Here, the worker writes the epw from memory to a file
            temp_epw, temp_dir = self.save_epw_to_temp()
            results = eplus.run_building(
                current_building,
                out_dir=out_dir,
                err_dir=err_dir,
                epw=temp_epw,
                error_mode=self.error_mode,
                ep_path=self.ep_path,
            )
            temp_dir.cleanup()
        else:
            results = eplus.run_building(
                current_building,
                out_dir=out_dir,
                err_dir=err_dir,
                epw=self.epw,
                error_mode=self.error_mode,
                ep_path=self.ep_path,
            )
        outputs = tuple((objective(results) for objective in self.problem.outputs))
        constraints = tuple(
            (constraint(results) for constraint in self.problem.constraints)
        )
        extracted_results = outputs + constraints
        if self.problem.add_outputs_list is not None:
            self.problem.record_results(values, results)
        self.update_pbar()
        if keep_dirs:
            return extracted_results + (out_dir,)
        else:
            return extracted_results

    @property
    def building(self):
        return self._building

    @building.setter
    def building(self, new_building) -> None:
        """Changes the building simulated.
        Changing this resets the cache.

        :param new_building: the building to use
        :return: None
        """
        for io in self.problem:
            io.setup(new_building)
        self.cache_clear()
        self._building = new_building

    @property
    def epw(self) -> PathLike:
        return self._epw

    @epw.setter
    def epw(self, value: PathLike) -> None:
        """Changes the epw file used with this building.
        Changing this resets the cache.

        :param value: path to the new epw file to use.
        :return: None
        """
        self.cache_clear()
        self._epw = value

    def generate_building(self, df: DF, index: int, file_name: str) -> None:
        """generate idf file

        :param df: dataFrame of the select row.
        :param index: start point.
        :param file_name: file name used to save as.
        :return: None
        """
        l = df.values.tolist()[index][: len(self.problem.inputs)]
        current_building = copy.deepcopy(self.building)
        for p, value in zip(self.problem.inputs, l):
            p.transformation_function(current_building, value)
        if isinstance(current_building, dict):
            with open(f"{file_name}.epJSON", "w") as fp:
                json.dump(current_building, fp)
        else:
            current_building.saveas(f"{file_name}.idf")


class EvaluatorEH(AbstractEvaluator, ABC):
    """This evaluator uses a Problem to modify an energy hub, and then solve it."""

    def __init__(
        self,
        problem: Problem,
        hub,
        out_dir: PathLike = config.out_dir,
        err_dir: PathLike = config.err_dir,
        error_mode: str = "Failfast",
        error_value: Optional[Sequence] = None,
        multi: bool = False,
        progress_bar: bool = True,
    ):
        """

        :param problem: a parametrization of the hub and the desired outputs
        :param hub: the energy hub that is being simulated.
        :param out_dir: the directory used for files created by the PyEHub simulation.
        :param err_dir: the directory where files from a failed run are stored.
        :param error_mode: One of {'Failfast', 'Silent', 'Print'}.
            Failfast: Any error aborts the evaluation.
            Silent: Evaluation will return the `error_value` for any lists of values that raise an error.
            Print: Same as silent, but warnings are printed to stderr for any errors.
        :param error_value: The value of the evaluation if an error occurs. Incompatible with error_mode='Failfast'.
        :param multi: whether or not to use multiprocessing, using this disables cache
        :param progress_bar: whether or not to display a progress bar
        """

        super().__init__(
            problem=problem,
            error_mode=error_mode,
            error_value=error_value,
            multi=multi,
            progress_bar=progress_bar,
        )
        # TODO: Either fully port the evaluator over to the config.py or switch everything over to config.yaml
        self.out_dir = out_dir
        self.err_dir = err_dir
        self.hub = hub
        self.problem = problem
        with open(Path(config.baseDir, "config.yaml"), "r") as file:
            self.config_settings = yaml.safe_load(file)

    # Need to overwrite __call__ due to default __call__ caching failing on time series values.
    def __call__(self, values: Sequence, **kwargs) -> tuple:
        # TODO: add caching that works with time series values
        return self._uncached_call(values, **kwargs)

    # override of validate due to pyehub model inputs being lists
    def validate(self, values: list) -> None:
        if len(values) != self.problem.num_inputs:
            raise ValueError(
                f"Wrong number of input values."
                f"{len(values)} provided, {self.problem.num_inputs} expected"
            )

    def eval_single(self, values: list) -> tuple:
        current_hub = copy.deepcopy(self.hub)
        pf.pyehub_parameter_editor(current_hub, self.problem.inputs, values)

        objectives = self.problem.names("outputs")
        current_hub.compile()
        primary_objective = objectives[0]
        current_hub.objective = primary_objective

        results = current_hub.solve()

        if self.out_dir is not None:
            output_file = self.config_settings["output_file"]
            self.out_dir = self.out_dir / output_file
            output_excel(
                results["solution"],
                self.out_dir,
                time_steps=len(current_hub.time),
                sheets=["other", "capacity_storage", "capacity_tech"],
            )

        outputs = tuple(
            (
                pf.get_by_path(results["solution"], [objective])
                for objective in objectives
            )
        )
        if self.problem.add_outputs_list is not None:
            self.problem.record_results(values, results)
        self.update_pbar()

        return outputs



import shutil

import numpy as np
import pytest

from besos.evaluator import EvaluatorGeneric
from besos.optimizer import rbf_opt
from besos.parameters import Parameter, RangeParameter
from besos.problem import Problem


@pytest.mark.skipif(shutil.which("bonmin") is None, reason="Requires bonmin to run.")
def test_rbfopt():
    """Simple test to make sure RBFOpt can be run properly"""

    def objective_function(x):
        return (x[0] * x[1] - x[2],)

    p1 = Parameter(value_descriptor=RangeParameter(1, 10), name="None_0")
    p2 = Parameter(value_descriptor=RangeParameter(0.1, 0.9), name="None_1")
    p3 = Parameter(value_descriptor=RangeParameter(0, 10), name="None_2")
    param_list = [p1, p2, p3]

    problem = Problem(param_list, 1)

    evaluator = EvaluatorGeneric(objective_function, problem, error_mode="Silent")

    opt = rbf_opt(evaluator, 30)
    value = opt.iloc[0]

    assert (
        np.isclose(value["None_0"], 1.0)
        and np.isclose(value["None_1"], 0.1)
        and np.isclose(value["None_2"], 10)
        and np.isclose(value["outputs_0"], -9.9)
    ), f"Unexpected output: {value}"  # None_1 assert values hardcoded to pass on Debian



import shutil
import subprocess
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from besos import config
from besos import eppy_funcs as ef
from besos import pyehub_funcs as pf
from besos import sampling
from besos.evaluator import EvaluatorEH, EvaluatorEP, EvaluatorGeneric
from besos.parameters import (
    FieldSelector,
    Parameter,
    ParameterEH,
    PathSelector,
    RangeParameter,
)
from besos.problem import EHProblem, EPProblem, Problem


needs_glpk = pytest.mark.skipif(
    shutil.which("glpsol") is None,
    reason=("Requires a PuLP compatible solver. Tests assume GLPK is the solver used."),
)


@pytest.fixture
def building():
    # returns the basic building
    return ef.get_building()


@pytest.fixture
def unexpanded():
    return ef.get_building(str(Path(config.data_dir, "unexpanded.idf")))


@pytest.fixture
def hub():
    # returns the basic hub
    return pf.get_hub()


@pytest.fixture
def problem():
    parameters = [
        Parameter(
            FieldSelector(
                object_name="Mass NonRes Wall Insulation", field_name="Thickness"
            )
        )
    ]
    objectives = [
        "Electricity:Facility",
        "Gas:Facility",
    ]  # the default is just 'Electricity:Facility'

    return EPProblem(parameters, objectives)


@pytest.fixture()
def sample_ready_problem():
    parameters = [
        Parameter(
            FieldSelector(
                object_name="Mass NonRes Wall Insulation", field_name="Thickness"
            ),
            RangeParameter(0.1, 0.9),
        )
    ]
    objectives = [
        "Electricity:Facility",
        "Gas:Facility",
    ]  # the default is just 'Electricity:Facility'

    return EPProblem(parameters, objectives)


@pytest.fixture()
def inputs(sample_ready_problem):
    return sampling.dist_sampler(sampling.lhs, sample_ready_problem, 2)


@pytest.fixture
def energyhub_df():
    return pd.DataFrame(np.array([[200, 600], [600, 200]]), columns=["p1", "p2"])


@pytest.fixture
def energyplus_df():
    return pd.DataFrame(np.array([[0.5], [0.8]]), columns=["p1"])


@pytest.fixture
def hub_problem():

    parameters = [
        Parameter(PathSelector(["LINEAR_CAPITAL_COSTS", "Boiler"])),
        Parameter(PathSelector(["LINEAR_CAPITAL_COSTS", "CHP"])),
    ]
    objectives = ["total_cost", "total_carbon"]

    return EHProblem(parameters, objectives)


@pytest.fixture
def oldstyle_hub_problem():
    parameters = []
    message = (
        "Use a :class:`PathSelector` and a :class:`Parameter` instead. "
        "These have been swapped in automatically."
    )
    with pytest.warns(DeprecationWarning, match=message):
        parameters.append(ParameterEH(["LINEAR_CAPITAL_COSTS", "Boiler"]))
    with pytest.warns(DeprecationWarning, match=message):
        parameters.append(ParameterEH(["LINEAR_CAPITAL_COSTS", "CHP"]))
    objectives = ["total_cost", "total_carbon"]

    return EHProblem(parameters, objectives)


@needs_glpk
@pytest.mark.compatibility
def test_equivalent_problems(hub_problem, oldstyle_hub_problem, hub):
    evaluator_new = EvaluatorEH(hub_problem, hub)
    evaluator_old = EvaluatorEH(oldstyle_hub_problem, hub)
    result_new = evaluator_new([200, 600])
    result_old = evaluator_old([200, 600])
    assert np.allclose(
        result_new, result_old
    ), "Hub compatibility transformation does not work"


@pytest.mark.slow
def test_evaluatorEP(building, problem, regtest):

    """To make sure EvaluatorEP can be initialised and works as intended"""

    evaluator = EvaluatorEP(problem, building)
    result = evaluator([0.5])  # run with thickness set to 0.5

    with regtest:
        print([f"{x:.5E}" for x in result])


@needs_glpk
def test_evaluatorEH_single(hub, hub_problem, regtest):
    """To make sure EvaluatorEH can be initialised and works as intended"""

    evaluator = EvaluatorEH(hub_problem, hub)
    result = evaluator([200, 600])

    with regtest:
        print([f"{x:.5E}" for x in result])


@needs_glpk
def test_evaluatorEH_df(hub, hub_problem, energyhub_df, regtest):
    """To make sure EvaluatorEH df_apply works as intended"""

    evaluator = EvaluatorEH(hub_problem, hub)
    result = evaluator.df_apply(energyhub_df)

    with regtest:
        print(result)


@needs_glpk
@pytest.mark.slow
def test_evaluatorEH_EP(hub, hub_problem, building, problem, regtest):
    """To make sure that base EvaluatorEP output can be used in an EvaluatorEH"""
    evaluatorEP = EvaluatorEP(problem, building)
    evaluatorEH = EvaluatorEH(hub_problem, hub)
    result = evaluatorEH(evaluatorEP([0.5]))

    with regtest:
        print([f"{x:.5E}" for x in result])


@needs_glpk
def test_evaluatorEH_TS(hub, regtest):
    """To make sure that EvaluatorEH can accept a time series as input"""
    timeseries = [
        {
            0: 4.0,
            1: 8.0,
            2: 6.0,
            3: 5.0,
            4: 7.0,
            5: 7.0,
            6: 7.0,
            7: 7.0,
            8: 7.0,
            9: 7.0,
            10: 7.0,
        }
    ]

    time_series_parameters = Parameter(PathSelector(["LOADS", "Elec"]))

    objectives = ["total_cost", "total_carbon"]
    time_series_problem = EHProblem(time_series_parameters, objectives)
    evaluatorEH = EvaluatorEH(time_series_problem, hub)
    result = evaluatorEH(timeseries)

    with regtest:
        print([f"{x:.5E}" for x in result])


@needs_glpk
def test_evaluatorEH_TS_df(hub, regtest):
    """To make sure that EvaluatorEH's df_apply can accept a dataframe of time series as input"""
    default_timeseries = [
        {
            0: 1.0,
            1: 4.0,
            2: 4.0,
            3: 4.0,
            4: 4.0,
            5: 4.0,
            6: 4.0,
            7: 4.0,
            8: 4.0,
            9: 4.0,
            10: 4.0,
        },
        {
            0: 20.0,
            1: 20.0,
            2: 20.0,
            3: 20.0,
            4: 20.0,
            5: 20.0,
            6: 20.0,
            7: 12.0,
            8: 12.0,
            9: 12.0,
            10: 12.0,
        },
    ]
    just_modified_heat = [
        {
            0: 1.0,
            1: 4.0,
            2: 4.0,
            3: 4.0,
            4: 4.0,
            5: 4.0,
            6: 4.0,
            7: 4.0,
            8: 4.0,
            9: 4.0,
            10: 4.0,
        },
        {
            0: 18.0,
            1: 18.0,
            2: 18.0,
            3: 18.0,
            4: 18.0,
            5: 18.0,
            6: 18.0,
            7: 16.0,
            8: 16.0,
            9: 16.0,
            10: 16.0,
        },
    ]
    just_modified_elec = [
        {
            0: 4.0,
            1: 8.0,
            2: 6.0,
            3: 5.0,
            4: 7.0,
            5: 7.0,
            6: 7.0,
            7: 7.0,
            8: 7.0,
            9: 7.0,
            10: 7.0,
        },
        {
            0: 20.0,
            1: 20.0,
            2: 20.0,
            3: 20.0,
            4: 20.0,
            5: 20.0,
            6: 20.0,
            7: 12.0,
            8: 12.0,
            9: 12.0,
            10: 12.0,
        },
    ]
    modified_timeseries = [
        {
            0: 4.0,
            1: 8.0,
            2: 6.0,
            3: 5.0,
            4: 7.0,
            5: 7.0,
            6: 7.0,
            7: 7.0,
            8: 7.0,
            9: 7.0,
            10: 7.0,
        },
        {
            0: 18.0,
            1: 18.0,
            2: 18.0,
            3: 18.0,
            4: 18.0,
            5: 18.0,
            6: 18.0,
            7: 16.0,
            8: 16.0,
            9: 16.0,
            10: 16.0,
        },
    ]
    timeseries_df = pd.DataFrame(
        np.array(
            [
                default_timeseries,
                just_modified_heat,
                just_modified_elec,
                modified_timeseries,
            ]
        ),
        columns=["p1", "p2"],
    )

    TSDFparameters = [
        Parameter(PathSelector(["LOADS", "Elec"])),
        Parameter(PathSelector(["LOADS", "Heat"])),
    ]
    TSDFobjectives = ["total_cost", "total_carbon"]
    TSDFproblem = EHProblem(TSDFparameters, TSDFobjectives)

    evaluatorEH = EvaluatorEH(TSDFproblem, hub)
    result = evaluatorEH.df_apply(timeseries_df)

    with regtest:
        print(result)


@needs_glpk
@pytest.mark.slow
def test_evaluatorEH_EP_df(hub, hub_problem, building, problem, energyplus_df, regtest):
    """To make sure that dataframe EvaluatorEP output can be used in an EvaluatorEH"""
    evaluatorEP = EvaluatorEP(problem, building)
    evaluatorEH = EvaluatorEH(hub_problem, hub)
    result = evaluatorEH.df_apply(evaluatorEP.df_apply(energyplus_df))

    with regtest:
        print(result)


def test_evaluatorGeneric(problem):
    """To make sure EvaluatorGeneric can be initialised and works as intended"""

    def function(values):
        return (values[0], values[0] ** 2)

    # this denotes a problem which takes 1 input, produces 2 outputs and no constraints.
    # The placeholder parameters/objectives will be generated automatically.
    new_problem = Problem(1, 2, 0)

    evaluator_1 = EvaluatorGeneric(function, problem)
    evaluator_2 = EvaluatorGeneric(function, new_problem)
    result_1 = evaluator_1([4])
    result_2 = evaluator_2([4])

    assert result_1 == (
        4,
        16,
    ), f"Unexpected result for EvaluatorGeneric with EPProblem, {result_1}"
    assert result_2 == (
        4,
        16,
    ), f"Unexpected result for EvaluatorGeneric with custom problem, {result_2}"


@pytest.mark.slow
def test_distributed_evaluatorEP(building, problem, regtest):
    """To make sure EvaluatorEP works with multi and distributed flags"""

    evaluator = EvaluatorEP(problem, building, multi=True, distributed=True)
    result = evaluator([0.5])  # run with thickness set to 0.5

    with regtest:
        print([f"{x:.5E}" for x in result])


@pytest.mark.slow
def test_error_evaluatorEP(problem):
    """To make sure EvaluatorEP error handling works"""
    building = ef.get_building(config.files.get("bad_idf"))
    evaluator = EvaluatorEP(problem, building)
    with pytest.raises(subprocess.CalledProcessError):
        with pytest.warns(UserWarning, match="problematic values were:"):
            evaluator([0.5])  # run with thickness set to 0.5


@pytest.mark.slow
def test_keep_output_evaluatorEP(building, sample_ready_problem, inputs, tmpdir):
    """Verify that the keep_output flag saves output in the correct location."""
    temp_path = tmpdir.mkdir("tmp")
    evaluator = EvaluatorEP(sample_ready_problem, building, out_dir=temp_path)
    df = evaluator.df_apply(inputs, keep_dirs=True)
    for dir_ in df["output_dir"]:
        # verify that the file exists and is readable
        with open(Path(dir_, "eplusout.dxf"), "r") as f:
            f.readline()


@pytest.fixture
def old_style_evaluation_function():
    def f(values):
        return (1.0, 2.0), ()

    return f


@pytest.mark.compatibility
def test_separate_outputs_and_constraints(old_style_evaluation_function):
    """Verify that we can still use the old evaluation function return format.

    Old format: (output1, output2, ...), (constraint1, constraint2, ...)
    New format: (output1, output2, ..., constraint1, constraint2, ...)

    """

    evaluator = EvaluatorGeneric(
        old_style_evaluation_function, Problem(inputs=1, outputs=2, constraints=0)
    )
    with pytest.warns(
        FutureWarning,
        match="Evaluators have changed the format used for storing outputs and constraints",
    ):
        assert evaluator([1]) == (1.0, 2.0)


@pytest.mark.compatibility
def test_epw_file_rename(building):
    """Verify that calls that use epw_file still work, but produce a warning."""
    with pytest.warns(FutureWarning, match="epw_file has been deprecated"):
        evaluator = EvaluatorEP(
            problem=Problem(), building=building, epw_file="test/path"
        )
    assert evaluator.epw == "test/path"


@pytest.mark.compatibility
def test_epw_duplicate_values(building):
    """Verify that calls using both epw and epw_file are rejected."""
    with pytest.raises(ValueError, match="epw_file and epw cannot be used together"):
        EvaluatorEP(problem=Problem(), building=building, epw="a", epw_file="b")


@pytest.mark.slow
def test_expand_objects(unexpanded):
    EvaluatorEP(problem=Problem(), building=unexpanded)([])



import pytest

from besos import eppy_funcs as ef
from errors import ModeError
from besos.parameters import (
    FieldSelector,
    FilterSelector,
    GenericSelector,
)


@pytest.fixture
def json_building():
    return ef.get_building(mode="json")


@pytest.fixture
def idf_building():
    return ef.get_building(mode="idf")


@pytest.fixture(params=["json", "idf"])
def building(request):
    return ef.get_building(mode=request.param)


@pytest.fixture
def filter_selector():
    def insulation_filter(building):
        mode = ef.get_mode(building)
        if mode == "idf":
            return [
                obj
                for obj in building.idfobjects["MATERIAL"]
                if "Insulation" in obj.Name
            ]
        elif mode == "json":
            return [
                obj
                for name, obj in building["Material"].items()
                if "Insulation" in name
            ]
        else:
            raise ModeError(mode)

    return FilterSelector(insulation_filter, "Thickness")


@pytest.fixture
def single_selector():
    return FieldSelector(
        class_name="Lights",
        object_name="Core_ZN_Lights",
        field_name="Watts per Zone Floor Area",
    )


@pytest.fixture
def intuited_selector():
    return FieldSelector(
        # class_name omitted
        object_name="1/2IN Gypsum",
        field_name="Conductivity",
    )


@pytest.fixture
def star_selector():
    return FieldSelector(
        class_name="Lights", object_name="*", field_name="Watts per Zone Floor Area"
    )


@pytest.fixture(
    params=["single_selector", "intuited_selector", "star_selector", "filter_selector"]
)
def selector(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def value():
    return 0.42


def test_init():
    """to make sure each type of selector is created as expected"""

    def insulation_filter(building):
        return [
            obj for name, obj in building["Material"].items() if "Insulation" in name
        ]

    field_selector = FieldSelector(
        class_name="Material",
        object_name="Mass NonRes Wall Insulation",
        field_name="Thickness",
    )
    filter_selector = FilterSelector(insulation_filter, "Thickness")
    generic_selector = GenericSelector(set=ef.wwr_all, setup=ef.one_window)

    # basic checks for the expected output format
    assert (
        str(field_selector)
        == "FieldSelector(field_name='Thickness', class_name='Material', "
        "object_name='Mass NonRes Wall Insulation')"
    ), f"FieldSelector not properly initialised: {field_selector}"
    assert (
        filter_selector.field_name == "Thickness"
        and str(type(filter_selector.get_objects)) == "<class 'method'>"
    ), f"FilterSelector not properly initialised: {filter_selector}"
    assert (
        str(type(generic_selector.set)) == "<class 'method'>"
        and str(type(generic_selector.setup)) == "<class 'method'>"
    ), f"GenericSelector not properly initialised: {generic_selector}"


def test_set_matches_get(building, selector, value):
    """Verifies that after setting a value with a selector, get returns the same value"""

    selector.set(building, value)

    assert set(selector.get(building)) == {
        value
    }, "Wrong value retrieved by get operation"


def test_idf_filter_selector(idf_building, filter_selector, value):
    building = idf_building
    filter_ = filter_selector

    found_objects = [obj.Name for obj in filter_.get_objects(building)]
    correct_objects = [
        "Mass NonRes Wall Insulation",
        "AtticFloor NonRes Insulation",
    ]

    assert (
        found_objects == correct_objects
    ), f"filter found {found_objects}, should have found {correct_objects}"

    filter_.set(building, value)
    assert (
        building.idfobjects["MATERIAL"][0].Thickness == value
    ), "Value set incorrectly"

    assert filter_.get(building) == [
        value,
        value,
    ], "Wrong value retrieved by get operation"


def test_json_filter_selector(json_building, filter_selector, value):
    building = json_building
    filter_ = filter_selector
    found_objects = filter_.get_objects(building)
    correct_objects = [
        building["Material"]["Mass NonRes Wall Insulation"],
        building["Material"]["AtticFloor NonRes Insulation"],
    ]
    assert (found_objects == correct_objects) or (
        found_objects == correct_objects[::-1]
    ), f"filter found {found_objects}, should have found {correct_objects}"

    filter_.set(building, value)

    assert (
        building["Material"]["AtticFloor NonRes Insulation"]["thickness"] == value
    ), "Value set incorrectly"



"""
This file contains custom errors used by BESOS
"""


class ModeError(ValueError):
    """An error for when an invalid mode is encountered."""

    def __init__(self, mode=None, message=None):
        if message is None:
            message = f'Invalid mode {mode}. Expected "idf" or "json"'
        super().__init__(message)



"""
The optimizer module provides wrappers which connect optimization packages to Evaluators.
Currently we support `platypus <https://github.com/Project-Platypus/Platypus>`_
and `rbf_opt <https://github.com/coin-or/rbfopt>`_

The supported Platypus algorithms are:

- GeneticAlgorithm
- EvolutionaryStrategy
- NSGAII
- EpsMOEA
- GDE3
- SPEA2
- MOEAD
- NSGAIII
- ParticleSwarm
- OMOPSO
- SMPSO
- CMAES
- IBEA
- PAES
- PESA2
- EpsNSGAII

These algorithms have the same configuration options as their counterparts in platypus,
with the parts provided by besos evaluators filled automatically.
"""

# Python Core Libraries
import os
import sys
from collections import OrderedDict
from functools import partial
from inspect import signature, Parameter, Signature
from typing import Type, List

# External Libraries
import dask_utils
import numpy as np
import pandas as pd
import platypus
import rbfopt

# BESOS Imports
from config import optimizer as conf
from evaluator import AbstractEvaluator
from problem import Problem


# TODO: evaluations or end criterion?
def platypus_alg(
    evaluator: AbstractEvaluator,
    algorithm: Type[platypus.Algorithm],
    evaluations: int = conf["evaluations"],
    *args,
    **kwargs,
) -> pd.DataFrame:
    """Uses a platypus algorithm to optimise over an evaluator
        If the evaluator is configured to run in parrallel (via 'multi' flag),
        the platypus algorithm will also run in parrallel.

    :param evaluator: An evaluation function to optimise over.
    :param algorithm: The platypus algorithm to use.
    :param evaluations: The algorithm will be stopped once it uses more than this many evaluations.
    :param args: arguments to pass to `algorithm`
    :param kwargs: keyword arguments to pass to `algorithm`.
    :return: the non-dominated solutions found by the algorithm.
    """

    if evaluator.multi:
        parrallel_platypus_evaluator = platypus.MapEvaluator(dask_utils.dask_map)
        kwargs.update({"evaluator": parrallel_platypus_evaluator})

    problem = evaluator.to_platypus()
    evaluator.problem.pre_optimisation()

    alg: platypus.AbstractGeneticAlgorithm = algorithm(problem, *args, **kwargs)
    alg.run(evaluations)
    # TODO: Have a smarter default for parts (instead of 'all')
    return solutions_to_df(platypus.unique(alg.result), evaluator.problem)


_ordered_argtypes = [
    (Parameter.POSITIONAL_OR_KEYWORD, False),
    (Parameter.POSITIONAL_OR_KEYWORD, True),
    (Parameter.VAR_POSITIONAL, False),
    (Parameter.KEYWORD_ONLY, False),
    (Parameter.KEYWORD_ONLY, True),
    (Parameter.VAR_KEYWORD, False),
]


def get_operator(problem: platypus.Problem, mutation=False):
    """Creates a valid operator for the given platypus Problem.
    This is provided to help when optimizing over problems with mixed types.

    :param problem: the platypus problem this operator should apply to
    :param mutation: if True create a mutation operator, if False create a variation operator
    :return:
    """
    operators = []
    if mutation:
        defaults = platypus.config.PlatypusConfig.default_mutator
        class_ = platypus.CompoundMutation
    else:
        defaults = platypus.config.PlatypusConfig.default_variator
        class_ = platypus.CompoundOperator
    for t in problem.types:
        if t in defaults:
            operators.append(defaults[t])
        else:
            for super_t, operator in defaults.items():
                if isinstance(t, super_t):
                    operators.append(operator)
    if mutation:
        return class_(*operators)
    return class_(*operators)


def _alg_t(algorithm):
    """A function that wraps a platypus algorithm in a common workflow.
    This function propagates the algorithm's function signature out to produce a more descriptive
    function signature."""

    # This is perhaps an overly fancy way of doing it, but it avoids having to
    # do the same process manually for all platypus algorithms, and will
    # update the signatures automatically if they change in platypus.

    defaults = {
        "variator": get_operator,
        "mutator": partial(get_operator, mutation=True),
    }

    def template(evaluator, evaluations=conf["evaluations"], *args, **kwargs):
        for kwarg, default in defaults.items():
            if kwarg not in kwargs or kwargs[kwarg] == "automatic":
                kwargs[kwarg] = default(evaluator.to_platypus())
        return platypus_alg(evaluator, algorithm, evaluations, *args, **kwargs)

    temp_name = (
        # (function, arguments_to_omit)
        (template, ["args", "kwargs"]),
        (algorithm.__init__, ["self", "problem"]),
    )

    argdict = {kind: [] for kind in _ordered_argtypes}
    for i, (func, omit) in enumerate(temp_name):
        sig = signature(func)
        p: Parameter
        for name, p in sig.parameters.items():
            if name in omit:
                continue
            if i > 0:
                # convert non-default keyword arguments to keyword only arguments
                if (
                    p.kind is Parameter.POSITIONAL_OR_KEYWORD
                    and p.default is Parameter.empty
                ):
                    p = p.replace(kind=Parameter.KEYWORD_ONLY)
                # use the defaults specified above
                if name in defaults:
                    p = p.replace(default="automatic")

            argdict[(p.kind, p.default is not Parameter.empty)].append(p)
    arglist = []
    for kind, default in _ordered_argtypes:
        arglist.extend(argdict[kind, default])

    template.__signature__ = Signature(
        parameters=arglist, return_annotation=signature(template).return_annotation
    )
    template.__name__ = algorithm.__name__
    return template


# The following functions are convenience wrappers for platypus' implementations
# of the algorithms with the same name
# they return the non-dominated solutions found by the algorithm

GeneticAlgorithm = _alg_t(platypus.GeneticAlgorithm)
EvolutionaryStrategy = _alg_t(platypus.EvolutionaryStrategy)
NSGAII = _alg_t(platypus.NSGAII)
EpsMOEA = _alg_t(platypus.EpsMOEA)
GDE3 = _alg_t(platypus.GDE3)
SPEA2 = _alg_t(platypus.SPEA2)
MOEAD = _alg_t(platypus.MOEAD)
NSGAIII = _alg_t(platypus.NSGAIII)
ParticleSwarm = _alg_t(platypus.ParticleSwarm)
OMOPSO = _alg_t(platypus.OMOPSO)
SMPSO = _alg_t(platypus.SMPSO)
CMAES = _alg_t(platypus.CMAES)
IBEA = _alg_t(platypus.IBEA)
PAES = _alg_t(platypus.PAES)
PESA2 = _alg_t(platypus.PESA2)
EpsNSGAII = _alg_t(platypus.EpsNSGAII)


def values_to_solution(
    values: list, problem: platypus.Problem, evaluate=True
) -> platypus.Solution:
    """Produces an platypus solution with the given inputs

    :param values: a list of input values to convert to a platypus solution
    :param problem: the platypus problem this solution applies to
    :param evaluate: whether to evaluate the solution before returning (solution will be unevaluated otherwise)
    :return: the corresponding platypus.Solution object
    """
    solution = platypus.Solution(problem)
    if len(values) != problem.nvars:
        raise ValueError(
            f"Length of values is {len(values)} but expected length of {problem.nvars}"
        )
    solution.variables = list(values)
    if evaluate:
        solution.evaluate()
    return solution


# TODO: Unify `parts` handling between this and Problem to remove duplicate code if possible


def solution_to_values(solution: platypus.Solution, parts="all") -> list:
    """Converts a platypus solution to a list containing the same values

    :param solution: a platypus solution to convert
    :param parts: which parts of the solution should be kept
    :return: a list of the requested values from the solution
    """
    part_options = OrderedDict(
        inputs="variables",
        outputs="objectives",
        constraints="constraints",
        violation="violation",
    )
    if parts == "all":
        parts = part_options.keys()
    assert set(parts) <= set(
        part_options
    ), f'parts must be a subset of {set(part_options.keys())} or "all"'
    result = []
    for part in parts:
        if part == "violation":
            result.append(solution.constraint_violation)
        elif part == "inputs":
            encoded = getattr(solution, part_options[part])
            decoded = [
                plat_type.decode(val)
                for val, plat_type in zip(encoded, solution.problem.types)
            ]
            result.extend(decoded)
        else:
            result.extend(getattr(solution, part_options[part]))
    return result


def solutions_to_df(
    solutions: List[platypus.Solution], problem, parts="all", flag_optimal=True
) -> pd.DataFrame:
    """Converts a list of platypus solutions to a DataFrame, with one row corresponding to each solution

    :param solutions: list of solutions to convert
    :param problem: the column names for DataFrame
    :param parts: which parts of the solutions should be kept
    :param flag_optimal: whether to include a boolean column denoting whether each solution is pareto-optimal
    :return: a DataFrame
    """

    def to_col_vals(solution_list):
        return list(
            zip(*(solution_to_values(solution, parts) for solution in solution_list))
        )

    solutions = platypus.unique(solutions)
    non_dominated = platypus.nondominated(solutions)
    columns = problem.names(parts)
    values, non_dom_vals = to_col_vals(solutions), to_col_vals(non_dominated)
    assert len(columns) == len(
        values
    ), f"{len(values)} values does not match {len(columns)} columns"
    # TODO: Intuit the dataframe column types based on the types of the parameters of the problem
    # or use the to_df method of the problem object
    solution_df = pd.DataFrame(
        {column: data for column, data in zip(columns, values)}
    )  # , dtype=float
    if flag_optimal:
        non_dom_df = pd.DataFrame(
            {column: data for column, data in zip(columns, non_dom_vals)}
        )  # , dtype=float
        df = pd.merge(solution_df, non_dom_df, how="outer", indicator="pareto-optimal")
        df["pareto-optimal"] = df["pareto-optimal"] == "both"
        if problem.depend_params or problem.add_outputs:
            df = problem.overwrite_df(df)
            problem.post_optimazation()
        return df
    if problem.depend_params or problem.add_outputs:
        solution_df = problem.overwrite_df(solution_df)
        problem.post_optimazation()
    return solution_df


def df_solution_to_solutions(
    df: pd.DataFrame,
    platypus_problem: platypus.Problem,
    besos_problem: Problem,
) -> List[platypus.Solution]:
    """Converts a solution DataFrame to a list of platypus solutions, with each row converted to one solution

    :param df: DataFrame to convert
    :param platypus_problem: platypus problem that the solutions apply to
    :param besos_problem: besos problem that the solutions apply to
    :return: a list of platypus solutions
    """
    input_df = df.copy(deep=True)
    labels = df.columns.values.tolist()
    for input in besos_problem.inputs:
        labels.remove(input.name)
    for label in labels:
        input_df.drop(columns=label, inplace=True)
    return list(input_df.apply(values_to_solution, problem=platypus_problem, axis=1))


def df_to_solutions(
    df: pd.DataFrame, problem: platypus.Problem
) -> List[platypus.Solution]:
    """Converts a DataFrame to a list of platypus solutions, with each row converted to one solution

    :param df: DataFrame to convert
    :param problem: platypus problem that the solutions apply to
    :return: a list of platypus solutions
    """
    return list(df.apply(values_to_solution, problem=problem, axis=1))


# rbf opt

# TODO: Add support for the other termination criterion
def rbf_opt(
    evaluator, evaluations, hide_output: bool = True, bonmin_path=None, rand_seed=None
):
    """
    This is wrapper to the rbfopt function.

    :param evaluator: An evaluation function to optimise over.
    :param evaluations: The algorithm will be stopped once it uses more than this many evaluations.
    :param hide_output: Whether to suppress output from the rbf_opt algorithm
    :param bonmin_path: The path to Bonmin, a specific feature useful for the cluster.
    :param rand_seed: Give a seed number to make sure the random has numbers.
    :return: a list of solutions

    """

    # default setting from rbfopt or do something custom (specifically for the cluster)
    # This parameter points rbfopt to the solver it was written to use.
    if bonmin_path == None:
        bonmin_path_set = "bonmin"
    else:
        bonmin_path_set = bonmin_path
    # This parameter seeds the random number generator of numpy.
    # To do a proper assessment this needs to be randomly set by the user.
    # else every run will just do the same search.
    if rand_seed == None:
        rand_seed_set = 937627691
    else:
        rand_seed_set = rand_seed

    problem = evaluator.problem
    inputs = problem.inputs
    dimension = problem.num_inputs
    num_outputs = problem.num_outputs

    assert (
        num_outputs == 1
    ), f"This is a single objective optimizer, but you have {num_outputs} objectives"
    assert problem.num_constraints == 0, "This optimizer cannot handle constraints"
    # may throw an attribute error if invalid parameters are used (i.e categorical)
    param_attrs = [
        (p.value_descriptor.min, p.value_descriptor.max, p.value_descriptor.rbf_type)
        for p in inputs
    ]
    var_lower, var_upper, var_type = (np.array(x) for x in zip(*param_attrs))
    # with the current classes var_type is just an array of 'R's
    # this would change if we add a descriptor for integers
    print(var_lower, var_upper, var_type)

    def wrapper(values):
        return evaluator(values)[0]

    def run_alg():
        bb = rbfopt.RbfoptUserBlackBox(
            dimension=dimension,
            var_lower=var_lower,
            var_upper=var_upper,
            var_type=var_type,
            obj_funct=wrapper,
        )
        settings = rbfopt.RbfoptSettings(
            max_evaluations=evaluations,
            minlp_solver_path=bonmin_path_set,
            rand_seed=rand_seed_set,
        )
        alg = rbfopt.RbfoptAlgorithm(settings, bb)
        return alg.optimize()

    if hide_output:
        with open(os.devnull, "w") as null:
            old, sys.stdout = sys.stdout, null
            try:
                val, x, itercount, evalcount, fast_evalcount = run_alg()
            finally:
                sys.stdout = old
    else:
        val, x, itercount, evalcount, fast_evalcount = run_alg()

    return problem.to_df(np.array([np.concatenate([x, [val]])]))



import pytest

from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorGeneric
from besos.parameters import (
    CategoryParameter,
    FieldSelector,
    Parameter,
    RangeParameter,
    wwr,
)
from besos.problem import Problem


# parameter - selector & descriptor
# selector - FilterSelector FieldSelector GenericSelector
# descriptor - RangeParameter CategoryParameter

# init, setup_checks, and setup_changes were moved directly from Will's code in
# parameters.py
def test_init():
    # inputs should be initialisable
    Parameter(
        FieldSelector(object_name="NonRes Fixed Assembly Window", field_name="UFactor"),
        value_descriptor=RangeParameter(min_val=0.1, max_val=5),
    )
    wwr()
    wwr(name="other")
    print("Parameters initialised")


def test_load_idf():
    idf = ef.get_idf()
    assert len(idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 21, "bad idf"


# TODO: look into fixing this
@pytest.mark.xfail
def test_setup_checks():
    # try excepts were updated to asserts and now the test fails, no exceptions are
    # being raised and it seems to be in 3rd party code

    idf = ef.get_idf()

    # eppy key-inputs should reject bad key values at setup
    r: Parameter = Parameter(FieldSelector(class_name="invalid", field_name="any"))

    with pytest.raises(Exception):
        r.setup(idf)
    print("Invalid object key detected successfully.")
    r = Parameter(
        FieldSelector(object_name="NonRes Fixed Assembly Window", field_name="invalid")
    )
    with pytest.raises(Exception):
        r.setup(idf)
    print("Invalid property detected successfully.")


def test_setup_changes():
    idf = ef.get_idf()

    r1 = Parameter(
        FieldSelector(object_name="NonRes Fixed Assembly Window", field_name="UFactor"),
        value_descriptor=RangeParameter(min_val=0.1, max_val=5),
    )
    r2 = wwr()
    r3 = wwr(name="other")

    r1.setup(idf)
    r2.setup(idf)
    assert len(idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 4, "bug in r2.setup"

    idf = ef.get_idf()
    assert len(idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 21, "bad idf"
    r3.setup(idf)
    assert len(idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]) == 4, "bug in r3.setup"

    print("setup tests done")


def test_custom_evaluation(regtest):
    """check to see if descriptors display as intended, and check to make sure custom
    evaluations work with EvaluatorGeneric"""

    # create the descriptors
    zero_to_nine = RangeParameter(min_val=0, max_val=9)
    single_digit_integers = CategoryParameter(options=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    text_example = CategoryParameter(options=["a", "b", "c", "other"])

    # create the parameters and the problem
    parameters = [
        Parameter(value_descriptor=zero_to_nine, name="0-9"),
        Parameter(value_descriptor=single_digit_integers, name="single digit"),
        Parameter(value_descriptor=text_example, name="text"),
    ]
    problem = Problem(parameters, outputs=["output"])

    # create the sampling distribution (seeded with only one output to give consistent
    # outputs after evaluating)
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, num_samples=1)
    with regtest:
        print(samples)

    # custom evaluation function from the jupyter notebook
    def evaluation_function(values):
        x, y, z = values
        if z == "other":
            return (0,)
        else:
            return (x * y,)

    evaluator = EvaluatorGeneric(evaluation_function, problem)
    # The evaluator will use this objective by default
    outputs = evaluator.df_apply(samples, keep_input=True)
    result = outputs.iloc[0]["output"]
    with regtest:
        print(f"{result:.5E}")



"""Objectives generate the output of evaluators.

For energyplus, outputs are calculated by reading the simulation output files,
using either MeterReaders or VariableReaders.

The optimisation direction (minimize or maximize) and if the objectives are used to put
constraints on a solution (such as electricity use being less than a certain amount)
are both specified by arguments to the Problem containing the objectives.

If the model used does not generate the needed output, the objective will modify the model
to add the corresponding output. For example, if a building does not include an
Ouptut:Meter object for Electricity:Facility, one will be added if the corresponding
:class:`MeterReader` is used with that building.
"""

# Python Core Libraries
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import List, Dict, Tuple, Union, Callable
from warnings import warn

# External Libraries
import pandas as pd
from packaging.version import Version

# BESOS Imports
import config
import eppy_funcs as ef
from IO_Objects import Objective
from errors import ModeError
from besostypes import PathLike


def _get_data_dict_regex():
    """Generates a regex expression that can parse a line of the data dictionary in the .mtr file
    Various assumptions are made about the exact composition of names."""
    report_code = r"(?P<report_code>\d+)"  # report code of digits
    num_items = r"(?P<num_items>\d+)"  # number of items composed of digits
    name = r"(?P<name>[^\[\]]+?)"  # assumes that name can be anything without '[]' characters, may be wrong
    units = r"\[(?P<units>[^\]]*?)\]"  # units are enclosed in square brackets.
    frequency = r"!(?P<frequency>[\w|,]+)"  # frequency preceded by a '!'
    remainder = r"(?:  ?\[(?P<remainder>[^\]]+)\])?"  # can be preceded by 1-2 spaces based on experience
    return re.compile(
        fr"^{report_code},{num_items},{name} {units} {frequency}{remainder}$"
    )


_values_collection = List[List[float]]


@dataclass
class EPResults:
    """Represents the information about a Meter or Variable read from a .eso file."""

    name: str
    frequency: str
    units: str
    data: pd.DataFrame

    # TODO: Consider improving the index for the DataFrame
    # index derived from the section headers (i.e. Design day winter hour=1) converted to a timestamp
    # this would simplify manipulating time series
    # Also consider nicer handling of duplicate names than just numbering

    @classmethod
    def from_pieces(
        cls, name, frequency, units, items, data, prefix=None
    ) -> "EPResults":
        if prefix is True:
            prefix = name
        unq_names = {}
        for i, original in enumerate(items):
            unq_names[original] = unq_names.get(original, []) + [i]
        new_names = items[:]
        for unq_name, values in unq_names.items():
            if len(values) == 1:
                continue  # not a duplicate, leave the name alone
            for count, index in enumerate(values):
                new_names[index] = f"{unq_name}_{count}"
        if prefix is not None:
            names = [f"{prefix}_{base_name}" for base_name in new_names]
        else:
            names = new_names
        df = pd.DataFrame({name: values for name, values in zip(names, zip(*data))})
        return cls(name, frequency, units, df)


_results_format = Dict[Tuple[str, str], EPResults]


def get_eso_version(eso: PathLike) -> Version:
    """Gets the version of energyplus used to generate an eso file.

    :param eso: the path to the eso file
    :return: the version of energyplus used to generate the eso file.
    """
    # a regular expression to extract information from the first line of an eso file
    first_line_regex = re.compile(
        r"^Program Version,EnergyPlus, Version "  # constant value at the start of the line
        r"(?P<version_number>[\d.]+)-(?P<hash>[0-9a-z]{10}), "
        r"YMD=(?P<date>\d\d\d\d\.\d\d\.\d\d+) (?P<time>\d?\d:\d\d)$",
    )
    with open(eso, "r") as f:
        first_line = f.readline()
    match = first_line_regex.match(first_line)
    if match is None:
        raise ValueError(
            f"First line of file:'{first_line}' did not match the expected format."
        )
    return Version(match.group("version_number"))


def read_eso(
    out_dir: PathLike = config.out_dir,
    file_name="eplusout.eso",
    version=None,
) -> _results_format:
    """Retrieve information from an Energy-Plus simulation, with outputs in out_dir

    :param out_dir: the directory that contains the output files
    :param file_name: the name of the output file to read from. Should be a .eso file.
    :param version: Deprecated argument. Do not use.
    :return: a dictionary with keys of the form {(MeterName, Reporting_Frequency or None):EPResults dataclass}
    """
    # output format described here:
    # https://energyplus.net/sites/default/files/pdfs_v8.3.0/OutputDetailsAndExamples.pdf

    eso_path = Path(out_dir, file_name)
    eso_version = get_eso_version(eso_path)
    with open(eso_path, "r") as file:
        f = file.readlines()  # may cause memory errors on very large files

    # Backwards compatibility warning.
    if version:
        warn(
            "The version number is now auto-detected."
            "The version argument is deprecated "
            "and will be removed in the future.",
            FutureWarning,
        )
        if Version(version) != eso_version:
            raise ValueError(
                f"Version {version} provided, "
                f"but eso file is from version {eso_version}"
            )
    del version

    # TODO: Find a way to avoid hardcoding these values to make things portable between versions
    # or move them to a version-dependant config of some kind
    if eso_version >= Version("9.0"):
        # header is always 5 lines for E+ 8.8 and 6 lines for E+ 9.0
        data_dictionary_header = 6
    else:
        data_dictionary_header = 5  # pragma: nocover

    data_dictionary_start = 1  # first line has already been processed
    data_dictionary_end = f.index("End of Data Dictionary\n")
    data_end = -2  # omit the End of Data line and the Number of Records Written line

    header = f[data_dictionary_start:data_dictionary_header]  # header currently ignored
    requested_vars = f[data_dictionary_header + 1 : data_dictionary_end]
    data = f[data_dictionary_end + 1 : data_end]

    data_dict_line = _get_data_dict_regex()
    codes = {}
    keys = set()

    for line in requested_vars:
        match = data_dict_line.search(line)
        if match is None:
            raise ValueError(f"could not match: {line}")
        name = match.group("name")
        frequency = match.group("frequency")
        items = match.group("remainder")
        items = ["Value"] if items is None else items.split(",")
        report_code = match.group("report_code")
        key = (name, frequency)
        assert report_code not in codes, f"already found report code {report_code}"
        assert key not in keys, f"already found item with {key}"

        result = dict(
            name=name,
            frequency=frequency,
            units=match.group("units"),
            items=items,
            data=[],
        )

        codes[report_code] = result
        keys.add(key)
    for line in data:
        report_code, *values = line.split(",")
        if report_code in codes:
            values = [float(value) for value in values]
            codes[report_code]["data"].append(values)

    return {
        (result["name"], result["frequency"]): EPResults.from_pieces(**result)
        for result in codes.values()
    }


def time_series_values(results: EPResults):
    """Returns the entire value column from the results.

    :param results: the results object
    :return: the values from the results object.
    """
    # TODO: Finish
    # Already is a pandas DF [to_dict() can be used to convert it to the
    # timeseries dict now or later in a different function?(maybe)]
    # possibly unit conversions as well? (although that should maybe
    # be done in a different function at a different time - would be easy if its a pandas dataframe)
    warn(
        "`time_series_values` is incomplete, and returns raw values, not time series values."
    )
    return results.data["Value"]


def sum_values(results: EPResults) -> float:
    """Returns the sum over the Value column from some results

    :param results: the results object to sum over
    :return: the sum of the first value for each entry in the collection.
    """
    return results.data["Value"].sum()


# TODO make "all" clear all optional output, currently incomplete (i.e. csv files)
def clear_outputs(building, outputs: Union[str, List[str]] = "all") -> None:
    """Disable certain types of output that appear in the .eso file.

    :param building: the building to modify
    :param outputs: Can be the class_name of the output to clear,
        a shortcut for different types of output,
        or a list that combines the above.
    :return: None
    """
    # `empty` is used so that different fields do not reference each the same container

    mode = ef.get_mode(building)

    if mode == "idf":
        building_items = building.idfobjects
        empty = list
    elif mode == "json":
        building_items = building
        empty = dict
    else:
        raise ModeError(mode)

    def find(prefix):
        return {
            x
            for x in building_items
            if x.startswith(ef.convert_format(prefix, "class", mode))
        }

    # These lists are probably incomplete!
    class_names = dict()
    # ENVIRONMENTALIMPACTFACTORS adds a bunch of meters, but is not a meter itself
    class_names["output_meters"] = find("Output:Meter:") | {
        "Output:Meter",
        "Output:EnvironmentalImpactFactors",
    }
    class_names["internal_meters"] = {"Meter:Custom", "Meter:CustomDecrement"}
    class_names["tables"] = find("Output:Table:")

    class_names["output"] = find("Output:")
    class_names["outputcontrol"] = find("OutputControl:")

    class_names["meters"] = (
        class_names["output_meters"] | class_names["internal_meters"]
    )

    if outputs == "all":
        outputs = chain(*class_names.values())
    else:
        if isinstance(outputs, str):  # make outputs a list
            outputs = [outputs]
        # convert each element in outputs via the class_names shortcuts
        outputs = set(chain(*(class_names.get(output, [output]) for output in outputs)))
    # clear the selected outputs
    for output in outputs:
        building_items[output] = empty()


class EPReader(Objective, ABC):
    field_pairs: tuple = NotImplemented

    def __init__(self, class_name, frequency: str = None, func=sum_values, name=""):
        super().__init__(name=name)
        self.class_name = class_name
        self.frequency = frequency
        # class name has a nonzero default, so this may be adding noise to the repr
        # but it does provide a more complete picture
        self._add_reprs(["class_name", "frequency"], check=True)

        self._process = func
        self._add_repr("func", "_process")

    def check_all(self, objective, mode):
        if mode == "idf":

            def get(attribute):
                return getattr(objective, attribute)

        elif mode == "json":

            def get(attribute):
                return objective[attribute]

        else:
            raise ModeError(mode)
        for self_attr, objective_attr in self.field_pairs:
            self_value = getattr(self, self_attr)
            objective_value = get(ef.convert_format(objective_attr, "field", mode))
            if self_value is not None and objective_value != self_value:
                return False
        return True

    def get_objective(self, building):
        """Retrieves output object in the building corresponding to this Objective.

        :param building: The building to search for the object.
        :return: The part of the building's data corresponding to this Objective.
            The return type varies between json and idf buildings
        """
        mode = ef.get_mode(building)

        if mode == "idf":
            objectives = building.idfobjects[
                ef.convert_format(self.class_name, "class", mode)
            ]
        elif mode == "json":
            objectives = building[
                ef.convert_format(self.class_name, "class", mode)
            ].values()
        else:
            raise ModeError(mode)
        for objective in objectives:
            if self.check_all(objective, mode):
                return objective
        raise ValueError(f"Cannot find the objective for {repr(self)}")

    def add_objective(self, building):
        """Creates and adds the meter needed by this objective to the building.

        :param building: the building to modify
        :return: None
        """
        try:
            self.get_objective(building)
        except ValueError:
            pass  # the objective is not present
        else:
            raise ValueError(f"Objective for {repr(self)} already exists")

        mode = ef.get_mode(building)

        new_object_dict = {
            ef.convert_format(field, "field", mode): getattr(self, attr)
            for attr, field in self.field_pairs
            if getattr(self, attr)
        }
        if self.frequency is None:
            new_object_dict[
                ef.convert_format("Reporting_Frequency", "field", mode)
            ] = "Hourly"
        if mode == "idf":
            building.newidfobject(
                key=ef.convert_format(self.class_name, "class", mode), **new_object_dict
            )
        elif mode == "json":
            # this is equivalent to appending, but e+ uses a dictionary instead of a list
            objectives = building[self.class_name]
            num = len(objectives)
            new_key = f"{self.class_name} {num}"
            assert (
                new_key not in building
            ), f"The building has incorrectly numbered {self.class_name} entries"
            objectives[new_key] = new_object_dict

    def validate(self, building):
        self.get_objective(building)

    def setup(self, building) -> None:
        """Prepares an idf so that it's output can be read by this meter.

        :param building: the idf to modify
        :return: None
        """
        try:
            self.get_objective(building)
        except ValueError:
            self.add_objective(building)
        self.validate(building)

    @abstractmethod
    def results_name(self):
        pass

    def __call__(self, results: _results_format) -> float:
        results_name = self.results_name()
        if self.frequency:
            meter_results = results[(results_name, self.frequency)]
        else:
            for (name, _), v in results.items():
                if name == results_name:
                    meter_results = v
                    break
            else:
                raise ValueError(f"No meter with name {results_name} found")
        return self._process(meter_results)


class MeterReader(EPReader):
    field_pairs = (("key_name", "Key_Name"), ("frequency", "Reporting_Frequency"))

    def __init__(
        self,
        key_name: str,
        class_name: str = config.objective_meter_type,
        frequency: str = None,
        func: Callable = sum_values,
        name: str = "",
    ):
        """

        :param key_name: What to measure (e.g. Electricity:Facility)
        :param class_name: The kind of meter to use (e.g. Output:Meter)
        :param frequency: the reporting frequency of the meter. (e.g. Timestep, Monthly)
        :param func: a function used to aggregate the results read from the file.
        :param name: the name of the output column corresponding to this objective
        """
        super().__init__(
            class_name=class_name, frequency=frequency, func=func, name=name
        )
        self.key_name = key_name
        self._add_repr("key_name", check=True)

    def results_name(self):
        return self.key_name

    @property
    def _default_name(self):
        return self.key_name


class VariableReader(EPReader):
    field_pairs = (
        ("key_value", "Key_Value"),
        ("variable_name", "Variable_Name"),
        ("frequency", "Reporting_Frequency"),
    )

    # TODO check if class_name must allays be Output:Variable
    def __init__(
        self,
        key_value,
        variable_name="*",
        class_name=config.objective_variable_type,
        frequency: str = None,
        func: Callable = sum_values,
        name="",
    ):
        """Reads the values of Variables from the .eso file
        Energyplus docs:
        https://bigladdersoftware.com/epx/docs/8-9/input-output-reference/input-for-output.html#outputvariable

        :param key_value: What to measure (e.g. Electricity:Facility)
        :param variable_name:  The kind of variable to use (e.g. Output:Variable)
        :param class_name: the type of variable to read from.
        :param frequency: the reporting frequency of the variable.
        :param func: a function used to aggregate the results read from the file.
        :param name: the name of the output column corresponding to this objective
        """
        super().__init__(
            class_name=class_name, frequency=frequency, func=func, name=name
        )
        self.key_value = key_value
        self._add_repr("key_value")
        self.variable_name = variable_name
        if self.variable_name != "*":
            self._add_repr("variable_name")

    def results_name(self):
        return f"{self.key_value},{self.variable_name}"

    @property
    def _default_name(self):
        return self.variable_name

    # FIXME: The return types for this function are inconsistent
    def __call__(self, results: _results_format) -> list:
        results_name = self.results_name()
        results_list = []

        if self.frequency:
            meter_results = results[(results_name, self.frequency)]
            return meter_results

        for (name, _), v in results.items():
            if self.key_value != "*":
                if name == results_name:
                    meter_results = v
                    return self._process(meter_results)
            else:
                if re.match(fr"\w+,{self.variable_name}", name):
                    meter_results = v
                    results_list.append(self._process(meter_results))

        if results_list:
            return results_list

        raise ValueError(f"No meter with name {results_name} found")


# TODO: Remove objectives from idf output automatically



"""
Functions related to under-the-hood interactions with energyplus.
"""

# Python Core Libraries
import json
import os
import platform
import shutil
import subprocess
import tempfile
import warnings
from pathlib import Path

# External Libraries
from deprecated.sphinx import deprecated

# BESOS Imports
import eppy_funcs as ef
import config
import objectives
from errors import ModeError
from besostypes import PathLike


# TODO: Make this a method of the building class when we add it
def get_idf_version(building):
    """ get energyplus version from idf or json file """
    mode = ef.get_mode(building)
    if mode == "idf":
        return building.idfobjects["VERSION"][0].Version_Identifier
    elif mode == "json":
        try:
            return building["Version"]["Version 1"]["version_identifier"]
        except KeyError:
            warnings.warn(f"Cannot find IDF version, substituting default.")
            return config.energy_plus_version
    else:
        raise ModeError(mode)


@deprecated(
    version="1.6.0",
    reason=("Version is derived automatically, this check is redundant."),
)
def check_idf_version(building, version):
    """ check if the version of energyplus matches the version in idf/json. """
    idf_version = get_idf_version(building).replace("-", ".")
    version = version.replace("-", ".")
    if idf_version != version[:3]:
        msg = f"IDF v{idf_version} does not match energyplus v{version[:3]}."
        warnings.warn(msg)


def has_hvac_templates(building) -> bool:
    """Returns whether or not the building contains HVACTemplate objects
    https://bigladdersoftware.com/epx/docs/8-0/input-output-reference/page-061.html

    :param building:
    :return: True if at leas one HVACTemplate object is present in the building.
    """
    mode = ef.get_mode(building)
    prefix = ef.convert_format("HVACTemplate", "class", mode)
    if mode == "idf":
        return any(
            k for k, v in building.idfobjects.items() if k.startswith(prefix) and v
        )
    else:
        return any(k for k in building if k.startswith(prefix))


def run_building(building, out_dir=config.out_dir, version=None, **eplus_args):
    """Run energy plus on a building object and return results

    If out_dir is not defined, the results will not be saved
    in the file system"""
    # backwards compatibility
    if version:
        warnings.warn(
            "the version argument is deprecated for run_building,"
            " and will be removed in the future",
            FutureWarning,
        )
        assert version == get_idf_version(building), "Incorrect version"

    with tempfile.TemporaryDirectory(dir=Path.home(), prefix=".besos_") as temp_dir:
        if out_dir is None:
            out_dir = temp_dir
        try:
            building_path = Path(temp_dir, "in.idf").resolve()
            building.saveas(str(building_path))
        except AttributeError:
            building_path = Path(temp_dir, "in.epJSON").resolve()
            with open(str(building_path), "w") as f:
                json.dump(building, f)
        expand_objects = has_hvac_templates(building)
        run_energyplus(
            building_path,
            out_dir=out_dir,
            version=get_idf_version(building),
            expand_objects=expand_objects,
            **eplus_args,
        )
        return objectives.read_eso(out_dir)


def run_energyplus(
    building_path: PathLike,
    out_dir: PathLike = config.out_dir,
    epw: PathLike = config.files["epw"],
    err_dir: PathLike = config.err_dir,
    schema_file=None,
    error_mode="Silent",
    version=config.energy_plus_version,
    ep_path=None,
    expand_objects: bool = False,
):
    """Run energy plus.
    This method is intended to work as similar to the cli tool as possible"""

    ep_exe_path, ep_directory = get_ep_path(version, ep_path)
    schema_file = schema_file or Path(ep_directory, "Energy+.idd")

    cmd = [
        ep_exe_path,
        "--idd",
        schema_file,
        "--weather",
        epw,
    ]
    if out_dir:
        cmd += ["--output-directory", out_dir]
    if expand_objects:
        cmd.append("--expandobjects")
    cmd.append(building_path)
    needs_shell = platform.system() == "Windows"
    try:
        subprocess.run(cmd, check=True, shell=needs_shell)
    except subprocess.CalledProcessError as e:
        # TODO: This log is excessively noisy. Can we cut it down to just the command's stderr?
        if error_mode != "Silent":
            # print eplus error
            filename = Path(out_dir, "eplusout.err")
            if os.path.exists(filename):
                err_file = open(filename, "r")
                for line in err_file:
                    print(line)
                print()
                err_file.close()
        if err_dir is not None and out_dir != err_dir:
            # copy eplus error files to err_dir
            if os.path.exists(err_dir):
                shutil.rmtree(err_dir)
            shutil.copytree(out_dir, err_dir)
        raise e


def get_ep_path(version, ep_path=None):
    """ get energyplus installation path by version"""
    if ep_path is not None:
        ep_directory = ep_path
        if platform.system() == "Windows":
            ep_exe = os.path.join(ep_directory, "energyplus.exe")
        elif platform.system() == "Linux":
            ep_exe = os.path.join(ep_directory, "energyplus")
        else:
            ep_exe = os.path.join(ep_directory, "energyplus")
    else:
        if len(version) == 3:
            if version == "9.0":
                version = "9-0-1"
            else:
                version = version.replace(".", "-") + "-0"
        else:
            version = version.replace(".", "-")
        # this is duplicated from eppy.runner.run_functions.paths_from_version
        if platform.system() == "Windows":
            ep_directory = "C:/EnergyPlusV{version}".format(version=version)
            ep_exe = os.path.join(ep_directory, "energyplus.exe")
        elif platform.system() == "Linux":
            ep_directory = "/usr/local/EnergyPlus-{version}".format(version=version)
            ep_exe = os.path.join(ep_directory, "energyplus")
        else:
            ep_directory = "/Applications/EnergyPlus-{version}".format(version=version)
            ep_exe = os.path.join(ep_directory, "energyplus")
    return ep_exe, ep_directory


def print_available_outputs(
    building,
    version=None,
    name=None,
    frequency=None,
):
    # backwards compatibility
    if version:
        warnings.warn(
            "the version argument is deprecated for print_available_outputs,"
            " and will be removed in the future",
            FutureWarning,
        )
        assert version == get_idf_version(building), "Incorrect version"

    if name is not None:
        name = name.lower()
    if frequency is not None:
        frequency = frequency.lower()
    results = run_building(building)
    for key in results.keys():
        if name is not None:
            if name not in key[0].lower():
                continue
            if frequency is not None:
                if key[1].lower() != frequency:
                    continue
        elif frequency is not None:
            if key[1].lower() != frequency:
                continue
        print(list(key))



from pathlib import Path

from eppy.modeleditor import IDDAlreadySetError
import pytest

from besos import config
from besos import eplus_funcs
from besos import eppy_funcs as ef


@pytest.fixture
def idd_default():
    _, ep_directory = eplus_funcs.get_ep_path(config.energy_plus_version)
    return Path(ep_directory, "Energy+.idd").resolve()


@pytest.fixture
def clean_idd_name():
    # reset the iddname
    # this kind of monkeypatch should not be used in production
    # it may mess with how idfs are processed

    # clear name before test
    old_name = ef.IDF.iddname
    ef.IDF.iddname = None
    yield
    # restore name after test
    ef.IDF.iddname = old_name


@pytest.fixture
def idd_other(clean_idd_name):
    return Path(config.data_dir, "Custom_Long_Fields.idd")


def test_idd_no_conflict_defaults(clean_idd_name):
    ef.get_idf()
    ef.get_idf()


def test_idd_no_conflict_explicit_default(idd_default, clean_idd_name):
    ef.get_idf()
    ef.get_idf(idd_file=idd_default)


def test_idd_no_conflict_same(idd_other, clean_idd_name):
    ef.get_idf(idd_file=idd_other)
    ef.get_idf(idd_file=idd_other)


def test_idd_conflict(idd_other, clean_idd_name):
    with pytest.raises(IDDAlreadySetError):
        ef.get_idf()
        ef.get_idf(idd_file=idd_other)


def test_idd_warns(idd_other, clean_idd_name):
    ef.get_idf(idd_file=idd_other)
    with pytest.warns(UserWarning, match="idd is already set to: "):
        ef.get_idf()



"""
Classes used to bundle the parameters, objectives and constraints,
and to manage operations that involve all of them at once,
such as converting data related to the problem to a DataFrame.
"""

# Python Core Libraries
import warnings
from typing import Union, List, Dict, Callable, Iterable

# External Libraries
import numpy as np
import pandas as pd
import platypus

# BESOS Imports
import IO_Objects
import config
import objectives
import parameters


# TODO: Consider storing the constraint bounds with the constraints themselves, not in the problem
# also consider storing the direction of optimisation inside the objectives
# might be able to inherit some of the constraint parsing from platypus, not sure if that is worth the hassle
class Problem(IO_Objects.ReprMixin):
    """
    A class that collects all of the inputs, outputs and constraints related to a model.
    Problems track what inputs are valid, and how to apply those inputs to a model.

    It tracks constraint bounds.
    Automatically converts certain shortcut notation:
    - Strings become name only Parameters
    - Integers become that many numbered Parameters
    Gives access to names of all parts of the problem
    Resolves duplicate names
    Converts numpy arrays to a DataFrame matching the format of the problem or some
    combination pieces of the format (ex: only inputs and constraints)
    Can convert to a platypus Problem (which lacks an evaluation function).
    """

    valid_parts: List[str] = ["inputs", "outputs", "constraints", "violation"]
    default_converters = {
        "outputs": IO_Objects.Objective,
        "constraints": IO_Objects.Objective,
    }

    def __init__(
        self,
        inputs: Union[int, List[Union[str, parameters.Parameter]]] = None,
        outputs: Union[int, List[Union[str, IO_Objects.Objective]]] = None,
        constraints: Union[int, List[Union[str, IO_Objects.Objective]]] = None,
        add_outputs: Union[int, List[Union[str, IO_Objects.Objective]]] = None,
        *,
        constraint_bounds: List[str] = None,
        minimize_outputs: List[bool] = None,
        converters: Dict[str, Callable[[str], IO_Objects.IOBase]] = None,
    ):
        """
        :param inputs: A list of Parameters, or an integer.
            If a list is used, strings are converted to Parameters and this list determines the valid inputs.
            If an integer, this problem accepts that many inputs.
        :param outputs: A list of Objectives, or an integer.
            If a list is used, strings are converted to Objectives and this list determines the valid inputs.
            If an integer, this problem requires that many outputs
        :param constraints: A list of Objectives to be used as constraints, or an integer.
            If a list is used, strings are converted to Objectives and this list determines the valid inputs.
            If an integer, this problem requires that many constraints.
        :param add_outputs: Outputs that don't need to be optimized in optimization algorithm
        :param constraint_bounds: A list of strings that are parsable by platypus, and which represent bounds
            on the corresponding constraints.
        :param minimize_outputs: A list of booleans indicating if the corresponding output should be
            minimized or maximized. (Default is to minimize all outputs).
        :param converters: A dictionary with keys from {"outputs", "constraints"} where values indicate how to
            convert those kinds of values to appropriate objectives/constraints for this problem.
        """
        super().__init__()
        self.converters = converters or self.default_converters
        extra_keys = set(self.converters.keys()) - set(self.valid_parts)
        if extra_keys:
            raise ValueError(
                f"The keys {extra_keys} are not valid for this Problem. Only {self.valid_parts} are valid"
            )

        self.inputs = self._io_to_list(inputs, "inputs")
        self.num_inputs = len(self.inputs)
        self.depend_params = {}

        self.outputs = self._io_to_list(outputs, "outputs")
        self.add_outputs = self._io_to_list(add_outputs, "outputs")
        self.add_outputs_list = None
        self.num_outputs = len(self.outputs)
        self.minimize_outputs = minimize_outputs or [True] * self.num_outputs
        msg = "outputs and minimize_outputs must have the same length"
        assert len(self.minimize_outputs) == self.num_outputs, msg

        self.constraints = self._io_to_list(constraints, "constraints")
        self.num_constraints = len(self.constraints)
        # TODO: consider using platypus's constraints here
        self.constraint_bounds = constraint_bounds or []
        msg = "constraints and constraint_bounds must have the same length"
        assert len(self.constraint_bounds) == self.num_constraints, msg
        self.fix_names()
        self._add_reprs(
            [
                "inputs",
                "outputs",
                "minimize_outputs",
                "constraints",
                "constraint_bounds",
                "converters",
            ],
            check=True,
        )

    def fix_names(self):
        mapping = {}
        duplicates = []
        for obj in self:
            mapping[obj.name] = mapping.get(obj.name, []) + [obj]
        for name, objects in mapping.items():
            if len(objects) != 1:
                duplicates.append((name, objects))
        duplicates_2 = []
        for name, objects in duplicates:
            edited_names = []
            for i, obj in enumerate(objects):
                try:  # try block in case of no selector initialized
                    obj.name = f"{obj.name}_{obj.selector.class_name}"
                except:
                    pass
                edited_names.append(obj.name)
                if edited_names.count(obj.name) > 1:
                    duplicates_2.append(obj)
        if duplicates_2:
            warnings.warn(
                RuntimeWarning(
                    f"Duplicate names found. (duplicate, repetitions): "
                    f"{[(name, len(objects)) for name, objects in duplicates]}"
                    f"\nAttempting to fix automatically"
                )
            )
        for i, obj in enumerate(duplicates_2):
            obj.name = f"{obj.name}_{(i+1)}"

    def _io_to_list(self, io_objects: Union[int, List[IO_Objects.IOBase], None], part):
        """Converts a list of objects to a standard form:
        numbered placeholders, original datatype or io_object that match the part provided.
        """
        if io_objects is None:
            return []
        if isinstance(io_objects, int):
            if part == "inputs":
                class_ = parameters.Parameter
            elif part in ["outputs", "constraints"]:
                class_ = IO_Objects.Objective
            else:
                raise ValueError(f"Cannot produce dummy values for part {part}")
            return [class_(name=f"{part}_{i}") for i in range(io_objects)]
        if isinstance(io_objects, (str, IO_Objects.IOBase)):
            io_objects = [io_objects]
        return [self.convert(o, part) for o in io_objects]

    def convert(self, io_object, part) -> IO_Objects.IOBase:
        """

        :param io_object: An object that should be converted to a parameter, objective or constraint
        :param part: one of 'inputs', 'outputs' or 'constraints' describing what to convert `io_object` to
        :return: the converted object
        """
        if isinstance(io_object, IO_Objects.IOBase):
            return io_object
        if part in self.converters:
            f = self.converters[part]
            try:
                return f(io_object)
            except TypeError as e:
                try:
                    if isinstance(io_object, dict):
                        return f(**io_object)
                    if isinstance(io_object, Iterable):
                        return f(*io_object)
                except:
                    pass
                raise TypeError(f"Cannot convert {io_object} to {part}") from e
        return io_object

    def expand_parts(self, parts: Union[str, List[str]]) -> List[str]:
        """Expands 'auto' and 'all' to the correct lists of parts, and wraps single parts in a list"""
        if parts == "auto":
            if self.num_constraints == 0:
                parts = ["inputs", "outputs"]
            else:
                parts = "all"
        if parts == "all":
            parts = self.valid_parts
        elif isinstance(parts, str):
            parts = [parts]

        if not set(parts) <= set(self.valid_parts):
            raise ValueError(
                f"parts must be a subset of {self.valid_parts + ['all']}, not {parts}"
            )
        return parts

    def names(self, parts: Union[str, List[str]] = "auto") -> List[str]:
        """

        :param parts: one of {'inputs', 'outputs', 'constraints', 'violation', 'all', 'auto'}
        :return: the names requested
        """
        parts = self.expand_parts(parts)
        names = []
        for attr in parts:
            if attr == "violation":
                names.append("violation")
            else:
                part = getattr(self, attr)
                if part is None:
                    raise ValueError(f"{attr} names not available")
                names.extend(IO_Objects.get_name(i) for i in part)
        return names

    # TODO: Add support for pareto-optimal column
    # TODO: Consolidate the different to_df code (ie from optimizer.py)
    def to_df(
        self,
        table: Union[np.array, pd.DataFrame],
        parts: Union[str, List[str]] = "auto",
    ) -> pd.DataFrame:
        """Converts the given table to a DataFrame that matches this problem's input/output format

        :param table: a table to be converted to a DataFrame. Must have the right number of columns.
        :param parts: inputs, outputs, constraints or all, depending on which data the DataFrame contains
        :return: A DataFrame containing the same data as the original table.
        """
        columns = self.names(parts)
        types = [
            p.pd_type if hasattr(p, "pd_type") else None
            for p in self.expand_parts(parts)
        ]
        if isinstance(table, pd.DataFrame):
            if len(table.columns) != len(columns):
                raise ValueError(
                    f"columns: {columns} requested but {list(table.columns)} found"
                )
            return table[columns]

        df = pd.DataFrame(table, columns=columns)
        # TODO: Make the categorical columns have the type category instead of object (attempt commented out below)
        # for col, type_ in zip(df, types):
        #     if type_:
        #         df[col] = df[col].astype(type_)
        return df

    def partial_df(self, table: Union[np.array, pd.DataFrame], parts="all"):
        parts = self.expand_parts(parts)
        for i in range(1, len(parts) + 1):
            partial_parts = parts[:i]
            try:
                return self.to_df(table, partial_parts), partial_parts
            except ValueError:
                continue
        raise ValueError("Could not find a matching DataFrame")

    def to_platypus(self) -> platypus.Problem:
        """Converts this problem to a platypus problem.
        No evaluator will be included.

        :return: A corresponding platypus problem
        """
        problem = platypus.Problem(
            self.num_inputs, self.num_outputs, self.num_constraints
        )
        for i, parameter in enumerate(self.inputs):
            problem.types[i] = parameter.platypus_type
        for i, direction in enumerate(self.minimize_outputs):
            problem.directions[i] = (
                platypus.Problem.MINIMIZE if direction else platypus.Problem.MAXIMIZE
            )
        for i, bound in enumerate(self.constraint_bounds):
            problem.constraints[i] = bound
        return problem

    def pre_optimisation(self):
        """Prepare optimisation for dependent parameter and non objectives."""
        for i in range(self.num_inputs):
            if type(self.inputs[i].value_descriptor).__name__ == "DependentParameter":
                self.depend_params[i] = []
        if self.add_outputs:
            self.add_outputs_list = []

    def overwrite_values(self, values):
        """Replace dataframe value with dependent parameter's value."""
        if not self.depend_params:
            return values
        l_values = list(values)
        for dp in self.depend_params:
            depend_value = self.get_depend_value(
                self.inputs[dp].value_descriptor, l_values
            )
            l_values[dp] = depend_value
            self.depend_params[dp].append(depend_value)
        return tuple(l_values)

    def get_depend_value(self, depend_param, values):
        """Return value of dependent parameter"""
        return depend_param.get_value(values[depend_param.index])

    def record_results(self, inputs, results):
        """Record add_outputs results"""
        inputs = list(inputs)
        for obj in self.add_outputs:
            inputs.append(obj(results))
        self.add_outputs_list.append(inputs)

    def get_non_objective(self, df):
        """Create a list for dataframe with add_outputs value"""
        l = []
        for row in df.values.tolist():
            for obj in self.add_outputs_list:
                for i in range(self.num_inputs):
                    if row[i] != obj[i]:
                        break
                    if i == self.num_inputs - 1:
                        for j in range(len(self.add_outputs)):
                            if len(l) < j + 1:
                                l.append([])
                            l[j].append(obj[i + j + 1])
        return l

    def overwrite_df(self, df):
        """Insert dependent parameter's value and/or add_outputs' value to dataframe"""
        labels = df.columns.values
        for index in self.depend_params:
            df.drop(columns=labels[index], inplace=True)
            df.insert(index, labels[index], self.depend_params[index])
        if self.add_outputs:
            l = self.get_non_objective(df)
            for i, data in enumerate(l):
                df.insert(
                    len(df.columns),
                    self.add_outputs[i].name,
                    data,
                )
        return df

    def post_optimazation(self):
        self.depend_params = {}
        self.add_outputs_list = None

    def __eq__(self, other):
        return (
            self.__class__ is other.__class__
            and self.inputs == other.inputs
            and self.outputs == other.outputs
            and self.constraints == other.constraints
        )

    def __iter__(self):
        return iter(self.inputs + self.outputs + self.constraints)


# TODO: consider having shortcuts for the converters instead of making this a whole different class
class EPProblem(Problem):
    """
    A problem with defaults that are appropriate for EnergyPlus simulations

    Strings for objectives/constraints become a MeterReader for the meter with
    that name.

    Integers still become numbered Parameters.
    """

    default_converters = {
        "outputs": objectives.MeterReader,
        "constraints": objectives.MeterReader,
    }

    def __init__(
        self,
        inputs=None,
        outputs=config.objectives,
        constraints=None,
        converters=None,
        **kwargs,
    ):
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            constraints=constraints,
            converters=converters,
            **kwargs,
        )


class EHProblem(Problem):
    """A problem that works with PyEHub models"""

    # TODO: Restructure if possible to be more like other problems.
    def __init__(
        self,
        inputs=None,
        outputs=["total_cost"],
        constraints=None,
        converters=None,
        **kwargs,
    ):
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            constraints=constraints,
            converters=converters,
            **kwargs,
        )

    # Overwritten functions to work with EvaluatorEH:
    def fix_names(self):
        pass



"""
Helper functions for interacting with PyEHub.
"""

# Python Core Libraries
import operator
from functools import reduce
from typing import List

# External Libraries
import numpy as np
import pandas as pd
import pyehub
from pyehub.energy_hub.ehub_model import EHubModel

# BESOS Imports
import config


"""This file is a collection of functions to interact with PyEHub."""

# TODO: replace back with yaml loader?
def get_xlsx(excel_file: str = config.files.get("xlsx")):
    return excel_file


def get_hub():
    """Generates the base PyEHub model from the excel file."""
    # could check the settings to see what type of ehub model to create
    data_file = get_xlsx()
    model = pyehub.energy_hub.EHubModel(excel=data_file)
    return model


def get_by_path(root, items):
    """Access a nested object in root by item sequence. Used to navigate PyEHub's not flat dict."""
    # TODO: inset try statement here iscase the parameter doesn't exist to fail gracefully
    return reduce(operator.getitem, items, root)


def set_by_path(root, items, value):
    """Set a value in a nested object in root by item sequence. Used to set values in PyEHub's not flat dict."""
    get_by_path(root, items[:-1])[items[-1]] = value


# no longer necessary due to restructure of EHProblem
def name_getter(dict):
    """Get list of names from a dictionary. Used to get the objective names for EHProblem."""
    namelist = []
    for item in dict:
        name = ""
        for part in item:
            if name == "":
                name = name + part
            else:
                name = name + ":" + part
        namelist.append(name)
    return namelist


# TODO: Is this really needed? Can we just use the transformation function?
def pyehub_parameter_editor(
    hub, parameters: List["parameters.Parameter"], values: list
):
    """Changes the __dict__ of the energy hub for the parameters specified
    on initialization with the values given to evaluator."""
    for (parameter, value) in zip(parameters, values):
        parameter.selector.set(hub, value)


def _split_EP_days(hourly_series):
    """
    Splits the hourly series of EP simulations that simulate multiple days in different seasons into specific energyhub time series.
    """
    # TODO: replace with a better loop of checking for the number of hours in the series and splitting better
    result1 = hourly_series.head(24)
    result1 = result1.reset_index()
    result2 = hourly_series.tail(24)
    result2 = result2.reset_index()
    return result1, result2

    # TODO:


def ep_to_eh(input, index_size):
    """
    Converts EnergyPlus evaluator outputs to PyEHub compatible inputs.
    """
    # TODO: Needs to be finished

    input_dict = input.to_dict()

    # Need to have a better joining on index for the pandas dataframe instead of appending
    # (difficulties to be noted with inserting dictionaries into cells)
    columnnames = []
    for j in input_dict:
        columnnames.append(j)

    # 2x for the number of seasons in the splitting will need to have some value stored from
    # the smarter splitting or make an easy spot to modify a value that propigates through the whole system.
    # if using the input for the function
    # index_size *=2
    # which should be something that does this
    # index_size = len(EPdf)*2
    # df_input = pd.DataFrame(np.nan, index=range(0,index_size), columns=columnnames)
    df_input = pd.DataFrame(columns=columnnames)

    for objective in input_dict:
        for series in input_dict[objective]:
            result = input_dict[objective][series].to_frame()
            # TODO: make a more genaric splitting process
            cold_result, warm_result = _split_EP_days(result)

            # TODO: make a more generic unit conversion function
            # 3.6 Megajoules per kWh
            cold_result = cold_result / 3600000
            warm_result = warm_result / 3600000
            cold_dict = cold_result.to_dict()
            cold_dict = cold_dict["Value"]
            warm_dict = warm_result.to_dict()
            warm_dict = warm_dict["Value"]

            cold_input = [cold_dict]
            warm_input = [warm_dict]

            temp_df1 = pd.DataFrame(np.array(cold_input), columns=[j])
            temp_df2 = pd.DataFrame(np.array(warm_input), columns=[j])
            # instead of appending I can edit the value of the specific cell with the input
            # (ah that gets difficult with dictionaries)
            # df_input.at[i, j] = [cold_dict]
            # df_input.iloc[i+len(EPdf), df_input.columns.get_loc(j)] = warm_dict
            # could do some sort of join on the
            df_input = df_input.append(temp_df1, ignore_index=True)
            df_input = df_input.append(temp_df2, ignore_index=True)

    return df_input



"""
This file is out of date, and is not maintained.

This file contains supporting code for the Buttons notebook.
"""

import os
from functools import partial

import hublib.ui as ui
import ipywidgets as w
from IPython.display import display

import config
import eppy_funcs as ef
import sampling


class FileManager:
    def __init__(self, extensions, dir=config.inDir, defaults=config.files, **kwargs):
        self.files = {}  # maps extension name to (current, {candidates}) tuples
        # defaults must already be in the directory
        for ext, file in defaults.items():
            if ext not in extensions:
                raise ValueError("{} is not a valid extension".format(ext))
            self.files[ext] = (file, {file})
        self.directory = dir
        self.extensions = extensions
        self.out = w.Output()
        self.displayOptions()

        widgetDefaults = {
            "name": "Files",
            "desc": "",
            "maxsize": "10M",
            "width": "30%",
            "maxnum": 5,
        }
        widgetDefaults.update(kwargs)

        multiLoader = ui.FileUpload(dir=dir, cb=self, **widgetDefaults)
        self.w = w.VBox([multiLoader.w, self.out])

    def __call__(self, w, fileNames):
        self.out.clear_output()
        badFiles = []
        with self.out:
            for name in fileNames:
                ext = name.split(".")[-1]
                name = self._short(name)
                if ext in self.extensions:
                    current, options = self.files.get(ext, ("File Needed", set()))
                    self.files[ext] = (name, options | {name})
                else:
                    badFiles.append(name)
            if badFiles:
                print(
                    " and ".join([", ".join(badFiles[:-1]), badFiles[-1]]),
                    "are invalid files. Only files with the extensions",
                    " or ".join([", ".join(self.extensions[:-1]), self.extensions[-1]]),
                    "are allowed",
                )
        self.displayOptions()
        w.reset()

    def displayOptions(self):
        with self.out:
            for ext in self.extensions:
                current, options = self.files.get(ext, ("File Needed", ["File Needed"]))
                dd = ui.Dropdown(
                    name=ext + " file",
                    description="",
                    value=current,
                    options=options,
                    cb=partial(self._setFile, ext=ext),
                )
                display(dd)

    def _short(self, name):
        return name[len(self.directory) + 1 :]

    def getFile(self, ext):
        # deliberetly does not catch key errors
        # might want to replace with __getitem__
        return os.path.join(self.directory, self.files[ext][0])

    def _setFile(self, w, file, ext):
        current, options = self.files[ext]
        self.files[ext] = (file, options)


fileTypes = ["building", "idd", "epw"]
fm = FileManager(fileTypes)


def output(transformers):
    idf = None
    results = w.Output()

    def _getSamples(w):
        nonlocal idf
        global df
        if idf is None:
            idf = ef.get_idf(
                fm.getFile("building"), fm.getFile("idd"), fm.getFile("epw")
            )
        with results:
            print("Simulating...")
            df = sampling.dist_sampler(
                samplerType.value, transformers, samplesBox.value
            )
            df["energy use"] = ef.Evaluator(transformers, idf).df_apply(df)
            print("Done")

            display(df)

    goButton = w.Button(description="Get Samples")
    goButton.on_click(_getSamples)
    return w.VBox([goButton, results])


startButtons = w.HBox([ui.HideCodeButton().w, ui.RunAllButton().w])

samplerType = ui.Dropdown(
    "Sampling method",
    "",
    options={
        "Latin Hypercube": sampling.lhs,
        "Random": sampling.random,
        "LHS with minimax criterion": sampling.lhs_maximin,
        "Random, seeded": sampling.seeded_sampler,
    },
    value=sampling.lhs,
)

samplesBox = ui.Integer("Samples", value=5, min=1, max=20)

interface = w.VBox([fm.w, samplerType.w, samplesBox.w])



"""
Dask related helper functions.
"""

# External Libraries
import dask.bag as db


def dask_map(f, l):
    dask_bag = db.from_sequence(l)
    map_result = db.map(f, dask_bag).compute()
    return map_result



import subprocess

import numpy as np
import pytest

from besos import config
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP, EvaluatorGeneric
from besos.parameters import CategoryParameter, FieldSelector, Parameter, RangeParameter
from besos.problem import EPProblem, Problem


@pytest.fixture
def building():
    return ef.get_building()


@pytest.fixture
def problem():
    problem = EPProblem(
        [
            Parameter(
                FieldSelector(
                    object_name="Mass NonRes Wall Insulation", field_name="Thickness"
                ),
                RangeParameter(min_val=0.01, max_val=0.99),
            ),
            Parameter(
                FieldSelector(
                    class_name="Construction",
                    object_name="ext-slab",
                    field_name="Outside Layer",
                ),
                CategoryParameter(options=("HW CONCRETE", "Invalid Material")),
            ),
        ]
    )
    return problem


@pytest.fixture
def samples(problem):
    # seeded_sampler results in indexes 0-3 being Invalid Material and 4 being HW
    # CONCRETE
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 5)
    return samples


def test_exception_throwing(building, problem, samples):
    """test to make sure exceptions are thrown in Failfast mode"""

    # check that an exception is raised with FailFast mode
    with pytest.raises(subprocess.CalledProcessError), pytest.warns(
        UserWarning, match="problematic values were:"
    ):
        EvaluatorEP(problem, building, error_mode="Failfast").df_apply(samples)


def test_exception_warn(building, problem, samples):
    # check that no exceptions are raised in the other mode
    with pytest.warns(UserWarning, match="problematic values were:"):
        EvaluatorEP(problem, building, error_mode="Print").df_apply(samples)


@pytest.mark.slow
def test_exception_silent(building, problem, samples):
    # check that no exceptions are raised in the other mode
    EvaluatorEP(problem, building, error_mode="Silent").df_apply(samples)


def test_error_bad_idf(problem):
    """To make sure EvaluatorEP error handling works"""
    building = ef.get_building(config.files.get("bad_idf"))
    evaluator = EvaluatorEP(problem, building)
    with pytest.raises(subprocess.CalledProcessError), pytest.warns(
        UserWarning, match="problematic values were:"
    ):
        evaluator([0.5, "HW CONCRETE"])


@pytest.fixture
def error_on_invalid():
    def f(values):
        print(values)
        if values[1] == "Invalid Material":
            print("invalid")
            raise ValueError
        else:
            return (1,)

    return f


def test_error_values(error_on_invalid, problem, samples):
    """check that automatic error handling will assign the desired error values"""

    # check that the default error value is assigned to invalid materials
    evaluator = EvaluatorGeneric(error_on_invalid, problem, error_mode="Silent")
    results = evaluator.df_apply(samples)
    value = results.iloc[0]["Electricity:Facility"]
    assert value == np.inf, (
        f"Invalid material not assigned the default error value, value assigned was: "
        f"{value}"
    )

    # check that a custom error value is assigned to invalid materials
    error_value = (-1,)
    evaluator = EvaluatorGeneric(
        error_on_invalid, problem, error_mode="Silent", error_value=error_value
    )
    results = evaluator.df_apply(samples)
    value = results.iloc[0]["Electricity:Facility"]
    assert value == -1, (
        f"Invalid material not assigned the correct error value, value assigned wa"
        f"s: {value}"
    )

    # check that valid inputs aren't assigned error values
    value = results.iloc[4]["Electricity:Facility"]
    assert value != -1, f"Valid material was assigned the error value: {error_value}"


@pytest.fixture
def error_value_tester():
    def helper(problem, error_value, **kwargs):
        return EvaluatorGeneric(
            lambda x: (1 / 0),
            problem,
            error_value=error_value,
            error_mode="Silent",
            **kwargs,
        )([])

    return helper


@pytest.mark.parametrize(
    "outputs, constraints, error_value, answer",
    [
        (0, 0, None, ()),
        (1, 0, None, (float("inf"),)),
        (
            2,
            0,
            None,
            (float("inf"), float("inf")),
        ),
        (
            2,
            0,
            (None, None),
            (float("inf"), float("inf")),
        ),
    ],
)
def test_error_objectives_both_versions(
    error_value_tester, outputs, constraints, error_value, answer
):
    """These tests should pass with both the new and the old error_value handling"""
    assert (
        error_value_tester(
            Problem(
                0,
                outputs,
                constraints,
                constraint_bounds=["dummy value"] * constraints,
            ),
            error_value=error_value,
        )
        == list(answer)
    )


@pytest.mark.compatibility
@pytest.mark.parametrize(
    "outputs, constraints, error_value, answer",
    [
        (0, 0, (None, None), ()),
        (1, 0, (None, None), (float("inf"),)),
        (
            3,
            0,
            (None, None),
            (
                float("inf"),
                float("inf"),
                float("inf"),
            ),
        ),
        (
            1,
            0,
            ((1,), ()),
            (1,),
        ),
        (
            2,
            0,
            ((1, 2), ()),
            (1, 2),
        ),
        (
            1,
            0,
            ((1,), None),
            (1,),
        ),
        (
            2,
            0,
            ((1, 2), None),
            (1, 2),
        ),
        (
            0,
            2,
            ((), (1, 2)),
            (1, 2),
        ),
        (
            0,
            2,
            (None, (1, 2)),
            (1, 2),
        ),
        (
            1,
            1,
            ((1,), (2,)),
            (1, 2),
        ),
        (
            1,
            1,
            (None, (1,)),
            (float("inf"), 1),
        ),
    ],
)
def test_error_objectives(
    error_value_tester, outputs, constraints, error_value, answer
):
    with pytest.warns(
        FutureWarning, match="Evaluators have changed the format used for error values."
    ):
        assert (
            error_value_tester(
                Problem(
                    0,
                    outputs,
                    constraints,
                    constraint_bounds=["dummy value"] * constraints,
                ),
                error_value=error_value,
            )
            == list(answer)
        )



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
with open("requirements.txt") as f:
    requirements = f.read().splitlines()
with open("requirements-complete.txt") as f:
    complete = f.read().splitlines()
with open("requirements-dev.txt") as f:
    dev = f.read().splitlines()

complete = complete[1:]  # Remove reference to requirements file
dev = dev[1:]  # Remove reference to requirement-complete file

setup(
    name="besos",
    version="1.7.1",
    description="A library for Building and Energy Simulation, Optimization and Surrogate-modelling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ralph Evins",
    author_email="revins@uvic.ca",
    url="https://gitlab.com/energyincities/besos",
    packages=["besos"],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    extras_require={
        "complete": complete,
        "dev": dev,
    },
)



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from sklearn import linear_model, pipeline
from sklearn.preprocessing import StandardScaler

from besos import config
from besos import eplus_funcs
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP
from besos.parameters import FieldSelector, Parameter, RangeParameter
from besos.problem import EPProblem


# Some code taken and adapted from example notebooks, made to work through pytest


@pytest.fixture
def building():
    return ef.get_building()


@pytest.fixture
def unexpanded():
    return ef.get_building(str(Path(config.data_dir, "unexpanded.idf").resolve()))


@pytest.fixture
def parameters():
    parameters = [
        Parameter(
            FieldSelector(
                object_name="NonRes Fixed Assembly Window",
                field_name="Solar Heat Gain Coefficient",
            ),
            value_descriptor=RangeParameter(0.01, 0.99),
        ),
        Parameter(
            FieldSelector("Lights", "*", "Watts per Zone Floor Area"),
            value_descriptor=RangeParameter(8, 12),
            name="Lights Watts/Area",
        ),
    ]
    return parameters


@pytest.fixture
def problem(parameters):
    objectives = ["Electricity:Facility"]
    problem = EPProblem(parameters, objectives)
    return problem


@pytest.fixture
def samples(problem):
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 10)
    return samples


@pytest.mark.slow
def test_expected_values(building, problem, samples, regtest):
    """check that the obtained results are consistent when using the same inputs"""

    def get_plot_data(model, density):
        # helper function from the example notebook
        p1 = problem.inputs[0].value_descriptor
        a = np.linspace(p1.min, p1.max, density)
        p2 = problem.inputs[1].value_descriptor
        b = np.linspace(p2.min, p2.max, density)
        plot_data = pd.DataFrame(
            np.transpose([np.tile(a, len(b)), np.repeat(b, len(a))]),
            columns=problem.names("inputs"),
        )
        return pd.concat([plot_data, pd.Series(model.predict(plot_data))], axis=1)

    evaluator = EvaluatorEP(problem, building, error_mode="Silent")
    train = evaluator.df_apply(samples, keep_input=True)
    print(problem.names())
    x, y, c = problem.names()

    # train and get the values
    model = pipeline.make_pipeline(StandardScaler(), linear_model.Ridge())
    model.fit(train[[x, y]].values, train[c].values)
    density = 30
    df = get_plot_data(model, int(density * 1.5))

    # check to see if the extremes and the midpoint are the expected values
    with regtest:
        for index in [0, 1012, 2024]:
            print(f"{df.iloc[index][0]:.5E}")


@pytest.mark.slow
def test_expand_objects_run(unexpanded):
    eplus_funcs.run_building(unexpanded, err_dir=config.err_dir / "BESOS_Errors")
